###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import pytest

from nucs.constants import STATS_IDX_SOLUTION_NB
from nucs.examples.queens.queens_problem import QueensDualProblem, QueensProblem
from nucs.heuristics.heuristics import VAR_HEURISTIC_FIRST_NOT_INSTANTIATED, VAR_HEURISTIC_SMALLEST_DOMAIN
from nucs.solvers.backtrack_solver import BacktrackSolver


class TestQueens:
    @pytest.mark.parametrize(
        "dual,var_heuristic,queen_nb,solution_nb",
        [
            (False, VAR_HEURISTIC_FIRST_NOT_INSTANTIATED, 1, 1),
            (False, VAR_HEURISTIC_FIRST_NOT_INSTANTIATED, 2, 0),
            (False, VAR_HEURISTIC_FIRST_NOT_INSTANTIATED, 3, 0),
            (False, VAR_HEURISTIC_FIRST_NOT_INSTANTIATED, 4, 2),
            (False, VAR_HEURISTIC_FIRST_NOT_INSTANTIATED, 5, 10),
            (False, VAR_HEURISTIC_FIRST_NOT_INSTANTIATED, 6, 4),
            (False, VAR_HEURISTIC_FIRST_NOT_INSTANTIATED, 7, 40),
            (False, VAR_HEURISTIC_FIRST_NOT_INSTANTIATED, 8, 92),
            (False, VAR_HEURISTIC_FIRST_NOT_INSTANTIATED, 9, 352),
            (False, VAR_HEURISTIC_SMALLEST_DOMAIN, 8, 92),
            (True, VAR_HEURISTIC_FIRST_NOT_INSTANTIATED, 1, 1),
            (True, VAR_HEURISTIC_FIRST_NOT_INSTANTIATED, 2, 0),
            (True, VAR_HEURISTIC_FIRST_NOT_INSTANTIATED, 3, 0),
            (True, VAR_HEURISTIC_FIRST_NOT_INSTANTIATED, 4, 2),
            (True, VAR_HEURISTIC_FIRST_NOT_INSTANTIATED, 5, 10),
            (True, VAR_HEURISTIC_FIRST_NOT_INSTANTIATED, 6, 4),
            (True, VAR_HEURISTIC_FIRST_NOT_INSTANTIATED, 7, 40),
            (True, VAR_HEURISTIC_FIRST_NOT_INSTANTIATED, 8, 92),
            (True, VAR_HEURISTIC_FIRST_NOT_INSTANTIATED, 9, 352),
            (True, VAR_HEURISTIC_SMALLEST_DOMAIN, 8, 92),
        ],
    )
    def test_queens_solve(self, dual: bool, var_heuristic: int, queen_nb: int, solution_nb: int) -> None:
        problem = QueensDualProblem(queen_nb) if dual else QueensProblem(queen_nb)
        solver = BacktrackSolver(problem, var_heuristic_idx=var_heuristic)
        solver.solve_all()
        assert solver.statistics[STATS_IDX_SOLUTION_NB] == solution_nb

    def test_queens_solve_2(self) -> None:
        problem = QueensProblem(2)
        solver = BacktrackSolver(problem)
        solver.find_one()
        assert solver.statistics[STATS_IDX_SOLUTION_NB] == 0



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from typing import List, Optional, Tuple, Union

import pytest

from nucs.constants import PROP_CONSISTENCY, PROP_ENTAILMENT, PROP_INCONSISTENCY
from nucs.propagators.element_l_eq_propagator import compute_domains_element_l_eq
from tests.propagators.propagator_test import PropagatorTest


class TestElementLEq(PropagatorTest):
    @pytest.mark.parametrize(
        "domains,parameters,consistency_result,expected_domains",
        [
            ([(-1, 0), (1, 2), (0, 2), (-1, 1)], [], PROP_CONSISTENCY, [[-1, 0], [1, 2], [0, 1], [-1, 1]]),
            ([(-4, -2), (1, 2), (0, 1), (0, 1)], [], PROP_ENTAILMENT, [[-4, -2], [1, 1], [1, 1], [1, 1]]),
            (
                [3, 0, 1, 2, 4, (0, 4), (-2, 0)],
                [0],
                PROP_ENTAILMENT,
                [[3, 3], [0, 0], [1, 1], [2, 2], [4, 4], [1, 1], [0, 0]],
            ),
            ([(-4, -2), (1, 2), (0, 1), (0, 0)], [], PROP_INCONSISTENCY, None),
        ],
    )
    def test_compute_domains(
        self,
        domains: List[Union[int, Tuple[int, int]]],
        parameters: List[int],
        consistency_result: int,
        expected_domains: Optional[List[List[int]]],
    ) -> None:
        self.assert_compute_domains(
            compute_domains_element_l_eq, domains, parameters, consistency_result, expected_domains
        )



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from nucs.constants import OPTIM_MODES, OPTIM_PRUNE
from nucs.examples.default_argument_parser import DefaultArgumentParser
from nucs.examples.golomb.golomb_problem import GolombProblem, golomb_consistency_algorithm
from nucs.solvers.backtrack_solver import BacktrackSolver
from nucs.solvers.consistency_algorithms import register_consistency_algorithm
from nucs.solvers.multiprocessing_solver import MultiprocessingSolver

# Run with the following command (the second run is much faster because the code has been compiled):
# NUMBA_CACHE_DIR=.numba/cache python -m nucs.examples.golomb -n 10 --symmetry_breaking
if __name__ == "__main__":
    parser = DefaultArgumentParser()
    parser.add_argument("-n", type=int, default=10)
    parser.add_argument("--opt_mode", choices=OPTIM_MODES, default=OPTIM_PRUNE)
    args = parser.parse_args()
    problem = GolombProblem(args.n, args.symmetry_breaking)
    consistency_alg_golomb = register_consistency_algorithm(golomb_consistency_algorithm)
    solver = (
        MultiprocessingSolver(
            [
                BacktrackSolver(
                    prob,
                    consistency_alg_idx=consistency_alg_golomb,
                    log_level=args.log_level,
                )
                for prob in problem.split(args.processors, 0)
            ]
        )
        if args.processors > 1
        else BacktrackSolver(
            problem,
            consistency_alg_idx=consistency_alg_golomb,
            log_level=args.log_level,
        )
    )
    solution = solver.minimize(problem.length_idx, mode=args.opt_mode)
    if args.stats:
        solver.print_statistics()
    if args.display:
        problem.print_solution(solution)



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import sys

from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import MAX, MIN
from nucs.heuristics.value_dom_heuristic import value_dom_heuristic


@njit(cache=True)
def min_cost_dom_heuristic(
    domains_stk: NDArray,
    not_entailed_propagators_stk: NDArray,
    dom_update_stk: NDArray,
    stks_top: NDArray,
    dom_idx: int,
    params: NDArray,
) -> int:
    """
    Chooses the value that minimizes the cost.
    :param domains_stk: the stack of shared domains
    :param not_entailed_propagators_stk: the stack of not entailed propagators
    :param dom_update_stk: the stack of domain updates
    :param stks_top: the index of the top of the stacks as a Numpy array
    :param dom_idx: the index of the shared domain
    :param params: a two-dimensional (first dimension corresponds to variables, second to values) costs array
    :return: the events
    """
    top = stks_top[0]
    best_cost = sys.maxsize
    best_value = -1
    shr_domain = domains_stk[top, dom_idx]
    for value in range(shr_domain[MIN], shr_domain[MAX] + 1):
        cost = params[dom_idx][value]
        if 0 < cost < best_cost:
            best_cost = cost
            best_value = value
    return value_dom_heuristic(
        domains_stk, not_entailed_propagators_stk, dom_update_stk, stks_top, dom_idx, best_value, params
    )



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from typing import List, Optional, Tuple, Union

import pytest

from nucs.constants import PROP_CONSISTENCY, PROP_ENTAILMENT
from nucs.propagators.affine_leq_propagator import compute_domains_affine_leq
from tests.propagators.propagator_test import PropagatorTest


class TestAffineLeq(PropagatorTest):
    @pytest.mark.parametrize(
        "domains,parameters,consistency_result,expected_domains",
        [
            ([(1, 10), (1, 10)], [1, -1, -1], PROP_CONSISTENCY, [[1, 9], [2, 10]]),
            ([(1, 10), (1, 10)], [1, 1, 8], PROP_CONSISTENCY, [[1, 7], [1, 7]]),
            ([(2, 3), (1, 2)], [1, 1, 5], PROP_ENTAILMENT, [[2, 3], [1, 2]]),
        ],
    )
    def test_compute_domains(
        self,
        domains: List[Union[int, Tuple[int, int]]],
        parameters: List[int],
        consistency_result: int,
        expected_domains: Optional[List[List[int]]],
    ) -> None:
        self.assert_compute_domains(
            compute_domains_affine_leq, domains, parameters, consistency_result, expected_domains
        )



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import copy
import logging
from typing import Any, List, Optional, Self, Sequence, Tuple, Union

import numpy as np
from numba import njit
from numpy.typing import NDArray
from rich import print

from nucs.constants import (
    EVENT_NB,
    NUMBA_DISABLE_JIT,
    PARAM,
    RANGE_END,
    RANGE_START,
    SIGNATURE_GET_TRIGGERS,
    TYPE_GET_TRIGGERS,
    VARIABLE,
)
from nucs.numba_helper import build_function_address_list, function_from_address
from nucs.propagators.propagators import GET_COMPLEXITY_FCTS, GET_TRIGGERS_FCTS

logger = logging.getLogger(__name__)


class Problem:
    """
    A problem is defined by:
    - a list of domains,
    - a list of variables and
    - a list of propagators.
    A variable references an offset and a domain.
    Domains can be shared between variables.
    """

    def __init__(
        self,
        domains: Union[List[Tuple[int, int]], List[int]],
        variables: Optional[List[int]] = None,
        offsets: Optional[List[int]] = None,
    ):
        """
        Initializes the problem.
        :param domains: the shared domains expressed as a list
        :param variables: the domain indices expressed as a list
        :param offsets: the domain offsets expressed as a list
        """
        n = len(domains)
        if variables is None:
            variables = list(range(0, n))
        if offsets is None:
            offsets = [0] * n
        self.domains = [[domain, domain] if isinstance(domain, int) else [domain[0], domain[1]] for domain in domains]
        self.variables = variables
        self.offsets = offsets
        self.domain_nb = len(domains)
        self.propagators: List[Tuple[List[int], int, List[int]]] = []
        self.propagator_nb = 0

    def split(self, split_nb: int, var_idx: int) -> List[Self]:
        """
        Splits a problem into several sub-problems by splitting the domain of a variable.
        :param split_nb: the number of sub-problems
        :param var_idx: the index of the variable
        :return: a list of sub-problems
        """
        logger.debug(f"Splitting in {split_nb} problems with variable {var_idx}")
        domain = self.domains[var_idx]
        domain_min = domain[0]
        domain_max = domain[1]
        domain_size = domain_max - domain_min + 1
        problems = []
        min_idx = domain_min
        for split_idx in range(split_nb):
            problem = copy.deepcopy(self)
            max_idx = min_idx + domain_size // split_nb - (0 if split_idx < domain_size % split_nb else 1)
            problem.domains[var_idx] = [min_idx, max_idx]
            min_idx = max_idx + 1
            problems.append(problem)
        return problems

    def add_variable(
        self, domain: Union[int, Tuple[int, int]], variable: Optional[int] = None, offset: Optional[int] = None
    ) -> int:
        """
        Adds an extra variable to the problem.
        :param domain: the shared domain of the variable
        :param variable: the domain index (automatically computed if not defined)
        :param offset: the domain offset (set to 0 if not defined)
        :return: the index of the extra variable
        """
        insertion_idx = len(self.domains)
        if variable is None:
            variable = insertion_idx
        if offset is None:
            offset = 0
        self.domains.append([domain, domain] if isinstance(domain, int) else [domain[0], domain[1]])
        self.variables.append(variable)
        self.offsets.append(offset)
        self.domain_nb = len(self.variables)
        return insertion_idx

    def add_variables(
        self,
        domains: Sequence[Union[int, Tuple[int, int]]],
        variables: Optional[List[int]] = None,
        offsets: Optional[List[int]] = None,
    ) -> int:
        """
        Adds extra variables to the problem.
        :param domains: the shared domains of the variables
        :param variables: the domain indices (automatically computed if not defined)
        :param offsets: the domain offsets (set to 0 if not defined)
        :return: the index where the extra variables have been added
        """
        insertion_idx = len(self.domains)
        n = len(domains)
        if variables is None:
            variables = [insertion_idx + i for i in range(n)]
        if offsets is None:
            offsets = [0] * n
        self.domains.extend(
            [[domain, domain] if isinstance(domain, int) else [domain[0], domain[1]] for domain in domains]
        )
        self.variables.extend(variables)
        self.offsets.extend(offsets)
        self.domain_nb = len(self.variables)
        return insertion_idx

    def add_propagator(self, propagator: Tuple[List[int], int, List[int]]) -> None:
        """
        Adds an extra propagator.
        :param propagator: the propagator
        """
        self.propagators.append(propagator)
        self.propagator_nb = len(self.propagators)

    def add_propagators(self, propagators: List[Tuple[List[int], int, List[int]]]) -> None:
        """
        Adds a list of propagators
        :param propagators: the propagators
        """
        self.propagators.extend(propagators)
        self.propagator_nb = len(self.propagators)

    def init(self) -> None:
        """
        Completes the initialization of the problem.
        """
        logger.debug("Initializing problem")
        # Sort the propagators based on their estimated amortized complexities.
        self.propagators.sort(key=lambda prop: GET_COMPLEXITY_FCTS[prop[1]](len(prop[0]), prop[2]))
        self.variables_arr = np.array(self.variables, dtype=np.uint32)
        self.offsets_arr = np.array(self.offsets, dtype=np.int32)
        self.algorithms = np.array([prop[1] for prop in self.propagators], dtype=np.uint8)
        # We will store propagator specific data in a global arrays, we need to compute variables and data bounds.
        logger.debug("Initializing bounds")
        self.bounds = np.zeros((max(1, self.propagator_nb), 2, 2), dtype=np.uint32)  # some redundancy here
        init_bounds(self.bounds, self.propagators)
        logger.debug("Initializing props")
        self.props_variables = np.empty(self.bounds[-1, VARIABLE, RANGE_END], dtype=np.uint32)
        self.props_offsets = np.empty(self.bounds[-1, VARIABLE, RANGE_END], dtype=np.int32)
        self.props_parameters = np.empty(self.bounds[-1, PARAM, RANGE_END], dtype=np.int32)
        init_props(
            self.props_variables,
            self.props_offsets,
            self.props_parameters,
            self.variables_arr,
            self.offsets_arr,
            self.bounds,
            self.propagators,
        )
        logger.debug("Initializing triggers")
        self.triggers = np.full((self.domain_nb, 1 << EVENT_NB, self.propagator_nb + 1), -1, dtype=np.int32)
        get_triggers_addrs = (
            np.empty(0)
            if NUMBA_DISABLE_JIT
            else np.array(build_function_address_list(GET_TRIGGERS_FCTS, SIGNATURE_GET_TRIGGERS))
        )
        init_triggers(
            self.triggers,
            self.domain_nb,
            self.propagator_nb,
            self.bounds,
            self.props_variables,
            self.props_parameters,
            self.algorithms,
            get_triggers_addrs,
        )
        logger.debug("Problem initialized")
        logger.info(f"Problem has {self.propagator_nb} propagators")
        logger.info(f"Problem has {self.domain_nb} variables")

    def solution_as_printable(self, solution: NDArray) -> Any:
        return solution.tolist()

    def print_solution(self, solution: Optional[NDArray]) -> None:
        if solution is not None:
            print(self.solution_as_printable(solution))
        else:
            print("No solution")


def init_bounds(bounds: NDArray, propagators: List[Tuple[List[int], int, List[int]]]) -> None:
    for prop_idx, prop in enumerate(propagators):
        if prop_idx > 0:
            bounds[prop_idx, :, RANGE_START] = bounds[prop_idx - 1, :, RANGE_END]
        bounds[prop_idx, VARIABLE, RANGE_END] = bounds[prop_idx, VARIABLE, RANGE_START] + len(prop[0])
        bounds[prop_idx, PARAM, RANGE_END] = bounds[prop_idx, PARAM, RANGE_START] + len(prop[2])


def init_props(
    props_variables: NDArray,
    props_offsets: NDArray,
    props_parameters: NDArray,
    variables_arr: NDArray,
    offsets_arr: NDArray,
    bounds: NDArray,
    propagators: List[Tuple[List[int], int, List[int]]],
) -> None:
    for prop_idx, prop in enumerate(propagators):
        var_start = bounds[prop_idx, VARIABLE, RANGE_START]
        var_end = bounds[prop_idx, VARIABLE, RANGE_END]
        props_variables[var_start:var_end] = variables_arr[prop[0]]  # cached for faster access
        props_offsets[var_start:var_end] = offsets_arr[prop[0]]  # cached for faster access
        param_start = bounds[prop_idx, PARAM, RANGE_START]
        param_end = bounds[prop_idx, PARAM, RANGE_END]
        props_parameters[param_start:param_end] = prop[2]


@njit(cache=True)
def init_triggers(
    triggers: NDArray,
    domain_nb: int,
    propagator_nb: int,
    bounds: NDArray,
    props_variables: NDArray,
    props_parameters: NDArray,
    algorithms: NDArray,
    get_triggers_addrs: NDArray,
) -> None:
    # for each domain and event, we store the list of propagator indices followed by -1
    indices = np.empty(domain_nb, dtype=np.int32)
    for event_mask in range(1, 1 << EVENT_NB):
        indices[:] = 0
        for prop_idx in range(propagator_nb):
            var_start = bounds[prop_idx, VARIABLE, RANGE_START]
            var_end = bounds[prop_idx, VARIABLE, RANGE_END]
            var_nb = var_end - var_start
            param_start = bounds[prop_idx, PARAM, RANGE_START]
            param_end = bounds[prop_idx, PARAM, RANGE_END]
            algorithm = algorithms[prop_idx]
            trigger_fct = (
                GET_TRIGGERS_FCTS[algorithm]
                if NUMBA_DISABLE_JIT
                else function_from_address(TYPE_GET_TRIGGERS, get_triggers_addrs[algorithm])
            )
            parameters = props_parameters[param_start:param_end]
            for var_idx in range(var_nb):
                if trigger_fct(var_nb, var_idx, parameters) & event_mask:
                    domain_idx = props_variables[var_idx + var_start]
                    triggers[domain_idx, event_mask, indices[domain_idx]] = prop_idx
                    indices[domain_idx] += 1



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import pytest

from nucs.constants import STATS_IDX_SOLUTION_NB
from nucs.examples.bibd.bibd_problem import BIBDProblem
from nucs.solvers.backtrack_solver import BacktrackSolver


class TestBIBD:
    @pytest.mark.parametrize(
        "v, b, r, k, l, solution_nb", [(6, 10, 5, 3, 2, 1), (7, 7, 3, 3, 1, 1), (8, 14, 7, 4, 3, 92)]
    )
    def test_bibd(self, v: int, b: int, r: int, k: int, l: int, solution_nb: int) -> None:
        problem = BIBDProblem(v, b, r, k, l)
        solver = BacktrackSolver(problem)
        solver.solve_all()
        assert solver.statistics[STATS_IDX_SOLUTION_NB] == solution_nb



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################

from nucs.examples.default_argument_parser import DefaultArgumentParser
from nucs.examples.donald.donald_problem import DonaldProblem
from nucs.heuristics.heuristics import DOM_HEURISTIC_MIN_VALUE, VAR_HEURISTIC_SMALLEST_DOMAIN
from nucs.solvers.backtrack_solver import BacktrackSolver

# Run with the following command (the second run is much faster because the code has been compiled):
# NUMBA_CACHE_DIR=.numba/cache python -m nucs.examples.donald
if __name__ == "__main__":
    parser = DefaultArgumentParser()
    args = parser.parse_args()
    problem = DonaldProblem()
    solver = BacktrackSolver(
        problem,
        var_heuristic_idx=VAR_HEURISTIC_SMALLEST_DOMAIN,
        dom_heuristic_idx=DOM_HEURISTIC_MIN_VALUE,
        log_level=args.log_level,
    )
    if args.all:
        solver.solve_all()
        solver.print_statistics()
    else:
        solution = solver.find_one()
        if args.stats:
            solver.print_statistics()
        if args.display:
            problem.print_solution(solution)



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from nucs.examples.employee_scheduling.employee_scheduling_problem import EmployeeSchedulingProblem


class TestEmployeeScheduling:
    def test_nurses(self) -> None:
        assert EmployeeSchedulingProblem().nurses(0, 0) == [0, 1, 2, 3, 4]
        assert EmployeeSchedulingProblem().nurses(0, 1) == [5, 6, 7, 8, 9]
        assert EmployeeSchedulingProblem().nurses(1, 0) == [15, 16, 17, 18, 19]
        assert EmployeeSchedulingProblem().nurses(1, 1) == [20, 21, 22, 23, 24]

    def test_shifts(self) -> None:
        assert EmployeeSchedulingProblem().shifts(0, 0) == [0, 5, 10]
        assert EmployeeSchedulingProblem().shifts(0, 1) == [1, 6, 11]
        assert EmployeeSchedulingProblem().shifts(1, 0) == [15, 20, 25]
        assert EmployeeSchedulingProblem().shifts(1, 1) == [16, 21, 26]



from nucs.problems.problem import Problem
from nucs.solvers.backtrack_solver import BacktrackSolver
from nucs.propagators.propagators import ALG_ALLDIFFERENT

n = 8  # the number of queens
problem = Problem(
    [(0, n - 1)] * n,  # these n domains are shared between the 3n variables with different offsets
    list(range(n)) * 3,  # for each variable, its shared domain
    [0] * n + list(range(n)) + list(range(0, -n, -1))  # for each variable, its offset
)
problem.add_propagator((list(range(n)), ALG_ALLDIFFERENT, []))
problem.add_propagator((list(range(n, 2 * n)), ALG_ALLDIFFERENT, []))
problem.add_propagator((list(range(2 * n, 3 * n)), ALG_ALLDIFFERENT, []))
print(BacktrackSolver(problem).find_one()[:n])


###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################


import numpy as np
from numba import njit
from numpy.typing import NDArray


def min_heap_init(capacity: int) -> NDArray:
    """
    Creates a data structure, designed to store integers from 0 to n-1,
    which is both a min-heap and a set (an element cannot be added twice).
    This data structure is implemented with a Numpy array of size 2n+1:
    the n first slots store the heap,
    the n following slots store the set,
    the last slot is the size of the heap.
    """
    return np.zeros(2 * capacity + 1, dtype=np.uint32)


@njit(cache=True)
def min_heap_add(heap: NDArray, capacity: int, val: int) -> None:
    if heap[capacity + val] == 0:
        heap[heap[-1]] = val
        min_heap_up(heap, heap[-1])
        heap[-1] += 1
        heap[capacity + val] = 1


@njit(cache=True)
def min_heap_pop(heap: NDArray, capacity: int) -> int:
    if heap[-1] == 0:
        return -1
    heap[-1] -= 1
    val = heap[0]
    min_heap_swap(heap, 0, heap[-1])
    min_heap_down(heap, 0)
    heap[capacity + val] = 0
    return val


@njit(cache=True)
def min_heap_swap(heap: NDArray, i: int, j: int) -> None:
    tmp = heap[i]
    heap[i] = heap[j]
    heap[j] = tmp


@njit(cache=True)
def min_heap_up(heap: NDArray, pos: int) -> None:
    """
    Non-recursive version because of a Numba bug.
    """
    while pos > 0:
        father = (pos - 1) >> 1
        if heap[father] <= heap[pos]:
            break
        min_heap_swap(heap, pos, father)
        pos = father


@njit(cache=True)
def min_heap_down(heap: NDArray, pos: int) -> None:
    """
    Non-recursive version because of a Numba bug.
    """
    while True:
        left = (pos << 1) + 1
        if left >= heap[-1]:
            break
        right = left + 1
        smallest = right if right < heap[-1] and heap[right] < heap[left] else left
        if heap[pos] <= heap[smallest]:
            break
        min_heap_swap(heap, pos, smallest)
        pos = smallest



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import (
    EVENT_MASK_MAX,
    EVENT_MASK_MIN,
    MAX,
    MIN,
    PROP_CONSISTENCY,
    PROP_ENTAILMENT,
    PROP_INCONSISTENCY,
)


def get_complexity_affine_geq(n: int, parameters: NDArray) -> float:
    """
    Returns the time complexity of the propagator as a float.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: a float
    """
    return n


@njit(cache=True)
def get_triggers_affine_geq(n: int, dom_idx: int, parameters: NDArray) -> int:
    """
    Returns the triggers for this propagator.
    :param parameters: the parameters
    :return: an array of triggers
    """
    return EVENT_MASK_MIN if parameters[dom_idx] < 0 else EVENT_MASK_MAX


@njit(cache=True)
def compute_domains_affine_geq(domains: NDArray, parameters: NDArray) -> int:
    """
    Implements Sigma_i a_i * x_i >= a_{n}.
    :param domains: the domains of the variables, x is an alias for domains
    :param parameters: the parameters of the propagator, a is an alias for parameters
    :return: the status of the propagation (consistency, inconsistency or entailment) as an int
    """
    domain_sum_min = domain_sum_max = -parameters[-1]
    factors = parameters[:-1]
    n = len(factors)
    for i in range(n):
        factor = factors[i]
        if factor > 0:
            domain_sum_min += factor * domains[i, MAX]
            domain_sum_max += factor * domains[i, MIN]
        elif factor < 0:
            domain_sum_min += factor * domains[i, MIN]
            domain_sum_max += factor * domains[i, MAX]
    if domain_sum_max >= 0:
        return PROP_ENTAILMENT
    for i in range(n):
        factor = factors[i]
        if factor != 0:
            if factor > 0:
                domains[i, MIN] = max(domains[i, MIN], domains[i, MAX] - (-domain_sum_min // -factor))
            else:
                domains[i, MAX] = min(domains[i, MAX], domains[i, MIN] + (domain_sum_min // -factor))
            if domains[i, MIN] > domains[i, MAX]:
                return PROP_INCONSISTENCY
    return PROP_CONSISTENCY



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import DOM_UPDATE_EVENTS, DOM_UPDATE_IDX, MAX, MIN, STATS_IDX_SOLVER_BACKTRACK_NB
from nucs.propagators.propagators import update_propagators


@njit(cache=True)
def cp_init(
    domains_stk: NDArray,
    not_entailed_propagators_stk: NDArray,
    dom_update_stk: NDArray,
    stks_top: NDArray,
    domains_arr: NDArray,
) -> None:
    """
    Initializes the choice points.
    :param domains_stk: the stack of shared domains
    :param not_entailed_propagators_stk: the stack of not entailed propagators
    :param dom_update_stk: the stack of domain updates
    :param stks_top: the index of the top of the stacks as a Numpy array
    :param domains_arr: the shared domains
    :return: a Numpy array
    """
    domains_stk[0] = domains_arr
    not_entailed_propagators_stk[0] = True
    dom_update_stk[:, :] = 0
    stks_top[0] = 0


@njit(cache=True)
def cp_put(domains_stk: NDArray, not_entailed_propagators_stk: NDArray, stks_top: NDArray) -> None:
    """
    Adds a choice point to the stack of choice points.
    :param domains_stk: the stack of shared domains
    :param not_entailed_propagators_stk: the stack of not entailed propagators
    :param stks_top: the index of the top of the stacks as a Numpy array
    """
    cp_top_idx = stks_top[0]
    domains_stk[cp_top_idx + 1, :, :] = domains_stk[cp_top_idx, :, :]
    not_entailed_propagators_stk[cp_top_idx + 1, :] = not_entailed_propagators_stk[cp_top_idx, :]
    stks_top[0] = cp_top_idx + 1


@njit(cache=True)
def backtrack(
    statistics: NDArray,
    not_entailed_propagators_stk: NDArray,
    dom_update_stk: NDArray,
    stks_top: NDArray,
    triggered_propagators: NDArray,
    triggers: NDArray,
) -> bool:
    """
    Backtracks and updates the problem's domains.
    :param statistics: the statistics array
    :param not_entailed_propagators_stk: the stack of not entailed propagators
    :param dom_update_stk: the stack of domain updates
    :param stks_top: the index of the top of the stacks as a Numpy array
    :param triggered_propagators: the set propagators that are currently triggered as a Numpy array
    :param triggers: a Numpy array of event masks indexed by shared domain indices and propagators
    :return: true iff it is possible to backtrack
    """
    if stks_top[0] == 0:
        return False
    stks_top[0] -= 1
    statistics[STATS_IDX_SOLVER_BACKTRACK_NB] += 1
    update_propagators(
        triggered_propagators,
        not_entailed_propagators_stk[stks_top[0]],
        triggers,
        dom_update_stk[stks_top[0], DOM_UPDATE_EVENTS],
        dom_update_stk[stks_top[0], DOM_UPDATE_IDX],
    )
    return True


@njit(cache=True)
def fix_choice_points(
    domains_stk: NDArray,
    stks_top: NDArray,
    variables_arr: NDArray,
    offsets_arr: NDArray,
    variable_idx: int,
    value: int,
    bound: int,
) -> bool:
    """
    Fixes the domain of the variable being optimized in the choice points.
    :param domains_stk: the stack of shared domains
    :param stks_top: the index of the top of the stacks as a Numpy array
    :param variables_arr: the variables as an array
    :param offsets_arr: the offsets as an array
    :param variable_idx: the index of the variable being optimized
    :param value: the current optimal value for the variable
    :param bound: the bound being optimized
    """
    dom_idx = variables_arr[variable_idx]
    domains_stk[0 : stks_top[0] + 1, dom_idx, bound] = value + (1 if bound == MIN else -1) - offsets_arr[variable_idx]
    while domains_stk[stks_top[0], dom_idx, MIN] > domains_stk[stks_top[0], dom_idx, MAX]:
        if stks_top[0] == 0:
            return False
        stks_top[0] -= 1
    return True


@njit(cache=True)
def fix_top_choice_point(
    domains_stk: NDArray,
    stks_top: NDArray,
    variables_arr: NDArray,
    offsets_arr: NDArray,
    variable_idx: int,
    value: int,
    bound: int,
) -> bool:
    """
    Fixes the domain of the variable being optimized in the top choice point.
    :param domains_stk: the stack of shared domains
    :param stks_top: the index of the top of the stacks as a Numpy array
    :param variables_arr: the variables as an array
    :param offsets_arr: the offsets as an array
    :param variable_idx: the index of the variable being optimized
    :param value: the current optimal value for the variable
    :param bound: the bound being optimized
    """
    top = stks_top[0]
    dom_idx = variables_arr[variable_idx]
    domains_stk[top, dom_idx, bound] = value + (1 if bound == MIN else -1) - offsets_arr[variable_idx]
    return domains_stk[top, dom_idx, MIN] <= domains_stk[top, dom_idx, MAX]



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import logging
from multiprocessing import Queue
from typing import Dict, Iterator, List, Optional, Tuple

import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import (
    LOG_LEVEL_INFO,
    MAX,
    MIN,
    NUMBA_DISABLE_JIT,
    OPTIM_RESET,
    PROBLEM_BOUND,
    PROBLEM_UNBOUND,
    SIGNATURE_COMPUTE_DOMAINS,
    SIGNATURE_CONSISTENCY_ALG,
    SIGNATURE_DOM_HEURISTIC,
    SIGNATURE_VAR_HEURISTIC,
    STATS_IDX_ALG_BC_NB,
    STATS_IDX_ALG_BC_WITH_SHAVING_NB,
    STATS_IDX_ALG_SHAVING_CHANGE_NB,
    STATS_IDX_ALG_SHAVING_NB,
    STATS_IDX_ALG_SHAVING_NO_CHANGE_NB,
    STATS_IDX_PROPAGATOR_ENTAILMENT_NB,
    STATS_IDX_PROPAGATOR_FILTER_NB,
    STATS_IDX_PROPAGATOR_FILTER_NO_CHANGE_NB,
    STATS_IDX_PROPAGATOR_INCONSISTENCY_NB,
    STATS_IDX_SOLUTION_NB,
    STATS_IDX_SOLVER_BACKTRACK_NB,
    STATS_IDX_SOLVER_CHOICE_DEPTH,
    STATS_IDX_SOLVER_CHOICE_NB,
    STATS_LBL_ALG_BC_NB,
    STATS_LBL_ALG_BC_WITH_SHAVING_NB,
    STATS_LBL_ALG_SHAVING_CHANGE_NB,
    STATS_LBL_ALG_SHAVING_NB,
    STATS_LBL_ALG_SHAVING_NO_CHANGE_NB,
    STATS_LBL_PROPAGATOR_ENTAILMENT_NB,
    STATS_LBL_PROPAGATOR_FILTER_NB,
    STATS_LBL_PROPAGATOR_FILTER_NO_CHANGE_NB,
    STATS_LBL_PROPAGATOR_INCONSISTENCY_NB,
    STATS_LBL_SOLUTION_NB,
    STATS_LBL_SOLVER_BACKTRACK_NB,
    STATS_LBL_SOLVER_CHOICE_DEPTH,
    STATS_LBL_SOLVER_CHOICE_NB,
    STATS_MAX,
    TYPE_CONSISTENCY_ALG,
    TYPE_DOM_HEURISTIC,
    TYPE_VAR_HEURISTIC,
)
from nucs.heaps import min_heap_init
from nucs.heuristics.heuristics import (
    DOM_HEURISTIC_FCTS,
    DOM_HEURISTIC_MIN_VALUE,
    VAR_HEURISTIC_FCTS,
    VAR_HEURISTIC_FIRST_NOT_INSTANTIATED,
)
from nucs.numba_helper import build_function_address_list, function_from_address
from nucs.problems.problem import Problem
from nucs.propagators.propagators import COMPUTE_DOMAINS_FCTS, reset_triggered_propagators, update_propagators
from nucs.solvers.choice_points import backtrack, cp_init, fix_choice_points, fix_top_choice_point
from nucs.solvers.consistency_algorithms import CONSISTENCY_ALG_BC, CONSISTENCY_ALG_FCTS
from nucs.solvers.queue_solver import QueueSolver
from nucs.solvers.solver import Solver, get_solution

logger = logging.getLogger(__name__)


class BacktrackSolver(Solver, QueueSolver):
    """
    A solver relying on a backtracking mechanism.
    """

    def __init__(
        self,
        problem: Problem,
        consistency_alg_idx: int = CONSISTENCY_ALG_BC,
        decision_domains: Optional[List[int]] = None,
        var_heuristic_idx: int = VAR_HEURISTIC_FIRST_NOT_INSTANTIATED,
        var_heuristic_params: List[List[int]] = [[]],
        dom_heuristic_idx: int = DOM_HEURISTIC_MIN_VALUE,
        dom_heuristic_params: List[List[int]] = [[]],
        stks_max_height: int = 512,
        log_level: str = LOG_LEVEL_INFO,
    ):
        """
        Initializes the solver.
        :param problem: the problem to be solved
        :param consistency_alg_idx: the index of the consistency algorithm
        :param decision_domains: the indices of the shared domains on which decisions will be made
        :param var_heuristic_idx: the index of the heuristic for selecting a variable/shared domain
        :param var_heuristic_params: a list of lists of parameters,
        usually parameters are costs and there is a list of value costs per variable/shared domain
        :param dom_heuristic_idx: the index of the heuristic for reducing a domain
        :param dom_heuristic_params: a list of lists of parameters,
        usually parameters are costs and there is a list of value costs per variable/shared domain
        :param stks_max_height: the maximal height of the choice point stacks
        :param log_level: the log level as a string
        """
        super().__init__(problem, log_level)
        decision_domains = list(range(problem.domain_nb)) if decision_domains is None else decision_domains
        logger.info(f"BacktrackSolver uses decision domains {decision_domains}")
        self.decision_domains = np.array(decision_domains, dtype=np.uint32)
        logger.info(f"BacktrackSolver uses variable heuristic {var_heuristic_idx}")
        self.var_heuristic_idx = var_heuristic_idx
        self.var_heuristic_params = np.array(var_heuristic_params, dtype=np.int64)
        logger.info(f"BacktrackSolver uses domain heuristic {dom_heuristic_idx}")
        self.dom_heuristic_idx = dom_heuristic_idx
        self.dom_heuristic_params = np.array(dom_heuristic_params, dtype=np.int64)
        logger.info(f"BacktrackSolver uses consistency algorithm {consistency_alg_idx}")
        self.consistency_alg_idx = consistency_alg_idx
        self.triggered_propagators = min_heap_init(problem.propagator_nb)
        reset_triggered_propagators(self.triggered_propagators, self.problem.propagator_nb)
        logger.debug("Initializing choice points")
        self.domains_stk = np.empty((stks_max_height, self.problem.domain_nb, 2), dtype=np.int32)
        self.not_entailed_propagators_stk = np.empty((stks_max_height, self.problem.propagator_nb), dtype=np.bool)
        self.dom_update_stk = np.empty((stks_max_height, 2), dtype=np.uint32)
        self.stks_top = np.ones((1,), dtype=np.uint16)
        logger.info(f"The stacks of the choice points have a maximal height of {stks_max_height}")
        cp_init(
            self.domains_stk,
            self.not_entailed_propagators_stk,
            self.dom_update_stk,
            self.stks_top,
            np.array(problem.domains),
        )
        logger.debug("Choice points initialized")
        logger.debug("Initializing statistics")
        self.statistics = np.array([0] * STATS_MAX, dtype=np.int64)
        logger.debug("Statistics initialized")
        logger.debug("BacktrackSolver initialized")

    def get_statistics_as_array(self) -> NDArray:
        return self.statistics

    def get_statistics_as_dictionary(self) -> Dict[str, int]:
        return {
            STATS_LBL_ALG_BC_NB: int(self.statistics[STATS_IDX_ALG_BC_NB]),
            STATS_LBL_ALG_BC_WITH_SHAVING_NB: int(self.statistics[STATS_IDX_ALG_BC_WITH_SHAVING_NB]),
            STATS_LBL_ALG_SHAVING_NB: int(self.statistics[STATS_IDX_ALG_SHAVING_NB]),
            STATS_LBL_ALG_SHAVING_CHANGE_NB: int(self.statistics[STATS_IDX_ALG_SHAVING_CHANGE_NB]),
            STATS_LBL_ALG_SHAVING_NO_CHANGE_NB: int(self.statistics[STATS_IDX_ALG_SHAVING_NO_CHANGE_NB]),
            STATS_LBL_PROPAGATOR_ENTAILMENT_NB: int(self.statistics[STATS_IDX_PROPAGATOR_ENTAILMENT_NB]),
            STATS_LBL_PROPAGATOR_FILTER_NB: int(self.statistics[STATS_IDX_PROPAGATOR_FILTER_NB]),
            STATS_LBL_PROPAGATOR_FILTER_NO_CHANGE_NB: int(self.statistics[STATS_IDX_PROPAGATOR_FILTER_NO_CHANGE_NB]),
            STATS_LBL_PROPAGATOR_INCONSISTENCY_NB: int(self.statistics[STATS_IDX_PROPAGATOR_INCONSISTENCY_NB]),
            STATS_LBL_SOLVER_BACKTRACK_NB: int(self.statistics[STATS_IDX_SOLVER_BACKTRACK_NB]),
            STATS_LBL_SOLVER_CHOICE_NB: int(self.statistics[STATS_IDX_SOLVER_CHOICE_NB]),
            STATS_LBL_SOLVER_CHOICE_DEPTH: int(self.statistics[STATS_IDX_SOLVER_CHOICE_DEPTH]),
            STATS_LBL_SOLUTION_NB: int(self.statistics[STATS_IDX_SOLUTION_NB]),
        }

    def minimize(self, variable_idx: int, mode: str = OPTIM_RESET) -> Optional[NDArray]:
        """
        Return the solution that minimizes a variable.
        :param variable_idx: the index of the variable to minimize
        :param mode: the optimization mode
        :return: the optimal solution if it exists or None
        """
        domain = self.domains_stk[self.stks_top[0], self.problem.variables_arr[variable_idx]]
        logger.info(f"Minimizing (mode {mode}) variable {variable_idx} (domain {domain}))")
        return self.optimize(variable_idx, MAX, mode)

    def maximize(self, variable_idx: int, mode: str = OPTIM_RESET) -> Optional[NDArray]:
        """
        Return the solution that maximizes a variable.
        :param variable_idx: the index of the variable to maximize
        :param mode: the optimization mode
        :return: the optimal solution if it exists or None
        """
        domain = self.domains_stk[self.stks_top[0], self.problem.variables_arr[variable_idx]]
        logger.info(f"Maximizing (mode {mode}) variable {variable_idx} (domain {domain}))")
        return self.optimize(variable_idx, MIN, mode)

    def optimize(self, variable_idx: int, bound: int, mode: str) -> Optional[NDArray]:
        """
        Finds, if it exists, the solution to the problem that optimizes a given variable.
        :param variable_idx: the index of the variable
        :param bound: the bound to optimize
        :param mode: the optimization mode
        :return: the solution if it exists or None
        """
        compute_domains_addrs, var_heuristic_addrs, dom_heuristic_addrs, consistency_alg_addrs = (
            get_function_addresses()
        )
        best_solution = None
        while (
            solution := solve_one(
                self.statistics,
                self.problem.algorithms,
                self.problem.bounds,
                self.problem.variables_arr,
                self.problem.offsets_arr,
                self.problem.props_variables,
                self.problem.props_offsets,
                self.problem.props_parameters,
                self.problem.triggers,
                self.domains_stk,
                self.not_entailed_propagators_stk,
                self.dom_update_stk,
                self.stks_top,
                self.triggered_propagators,
                self.consistency_alg_idx,
                self.decision_domains,
                self.var_heuristic_idx,
                self.var_heuristic_params,
                self.dom_heuristic_idx,
                self.dom_heuristic_params,
                compute_domains_addrs,
                consistency_alg_addrs,
                var_heuristic_addrs,
                dom_heuristic_addrs,
            )
        ) is not None:
            logger.info(f"Found a local optimum: {solution[variable_idx]}")
            best_solution = solution
            if mode == OPTIM_RESET:
                logger.debug("Resetting solver")
                cp_init(
                    self.domains_stk,
                    self.not_entailed_propagators_stk,
                    self.dom_update_stk,
                    self.stks_top,
                    np.array(self.problem.domains),
                )
                if not fix_top_choice_point(
                    self.domains_stk,
                    self.stks_top,
                    self.problem.variables_arr,
                    self.problem.offsets_arr,
                    variable_idx,
                    best_solution[variable_idx],
                    bound,
                ):
                    break
            else:
                logger.debug("Pruning choice points")
                if self.stks_top[0] == 0:
                    break
                self.stks_top[0] -= 1
                if not fix_choice_points(
                    self.domains_stk,
                    self.stks_top,
                    self.problem.variables_arr,
                    self.problem.offsets_arr,
                    variable_idx,
                    best_solution[variable_idx],
                    bound,
                ):
                    break
            reset_triggered_propagators(self.triggered_propagators, self.problem.propagator_nb)
        return best_solution

    def solve(self) -> Iterator[NDArray]:
        """
        Returns an iterator over the solutions.
        :return: an iterator
        """
        logger.info("Solving and iterating over the solutions")
        compute_domains_addrs, var_heuristic_addrs, dom_heuristic_addrs, consistency_alg_addrs = (
            get_function_addresses()
        )
        while True:
            solution = solve_one(
                self.statistics,
                self.problem.algorithms,
                self.problem.bounds,
                self.problem.variables_arr,
                self.problem.offsets_arr,
                self.problem.props_variables,
                self.problem.props_offsets,
                self.problem.props_parameters,
                self.problem.triggers,
                self.domains_stk,
                self.not_entailed_propagators_stk,
                self.dom_update_stk,
                self.stks_top,
                self.triggered_propagators,
                self.consistency_alg_idx,
                self.decision_domains,
                self.var_heuristic_idx,
                self.var_heuristic_params,
                self.dom_heuristic_idx,
                self.dom_heuristic_params,
                compute_domains_addrs,
                consistency_alg_addrs,
                var_heuristic_addrs,
                dom_heuristic_addrs,
            )
            if solution is None:
                break
            logger.debug("Found a solution")
            yield solution
            if not backtrack(
                self.statistics,
                self.not_entailed_propagators_stk,
                self.dom_update_stk,
                self.stks_top,
                self.triggered_propagators,
                self.problem.triggers,
            ):
                break

    def minimize_and_queue(self, variable_idx: int, processor_idx: int, solution_queue: Queue, mode: str) -> None:
        """
        Enqueues the solution that minimizes a variable.
        :param variable_idx: the index of the variable to minimize
        :param processor_idx: the index of the processor running the minimizer
        :param solution_queue: the solution queue
        :param mode: the optimization mode
        """
        domain = self.domains_stk[self.stks_top[0], self.problem.variables_arr[variable_idx]]
        logger.info(f"Minimizing (mode {mode}) variable {variable_idx} (domain {domain})) and queuing solutions")
        self.optimize_and_queue(variable_idx, MAX, processor_idx, solution_queue, mode)

    def maximize_and_queue(self, variable_idx: int, processor_idx: int, solution_queue: Queue, mode: str) -> None:
        """
        Enqueues the solution that maximizes a variable.
        :param variable_idx: the index of the variable to maximizer
        :param processor_idx: the index of the processor running the maximizer
        :param solution_queue: the solution queue
        :param mode: the optimization mode
        """
        domain = self.domains_stk[self.stks_top[0], self.problem.variables_arr[variable_idx]]
        logger.info(f"Minimizing (mode {mode}) variable {variable_idx} (domain {domain})) and queuing solutions")
        self.optimize_and_queue(variable_idx, MIN, processor_idx, solution_queue, mode)

    def optimize_and_queue(
        self, variable_idx: int, bound: int, processor_idx: int, solution_queue: Queue, mode: str
    ) -> None:
        """
        Enqueues the solution that optimizes a variable.
        :param variable_idx: the index of the variable
        :param bound: the bound to optimize
        :param processor_idx: the index of the processor
        :param solution_queue: the solution queue
        :param mode: the optimization mode
        """
        compute_domains_addrs, var_heuristic_addrs, dom_heuristic_addrs, consistency_alg_addrs = (
            get_function_addresses()
        )
        while True:
            solution = solve_one(
                self.statistics,
                self.problem.algorithms,
                self.problem.bounds,
                self.problem.variables_arr,
                self.problem.offsets_arr,
                self.problem.props_variables,
                self.problem.props_offsets,
                self.problem.props_parameters,
                self.problem.triggers,
                self.domains_stk,
                self.not_entailed_propagators_stk,
                self.dom_update_stk,
                self.stks_top,
                self.triggered_propagators,
                self.consistency_alg_idx,
                self.decision_domains,
                self.var_heuristic_idx,
                self.var_heuristic_params,
                self.dom_heuristic_idx,
                self.dom_heuristic_params,
                compute_domains_addrs,
                consistency_alg_addrs,
                var_heuristic_addrs,
                dom_heuristic_addrs,
            )
            if solution is None:
                break
            logger.info(f"Found a local optimum: {solution[variable_idx]}")
            logger.info(self.stks_top[0])
            solution_queue.put((processor_idx, solution, self.statistics))
            if mode == OPTIM_RESET:
                logger.debug("Resetting solver")
                cp_init(
                    self.domains_stk,
                    self.not_entailed_propagators_stk,
                    self.dom_update_stk,
                    self.stks_top,
                    np.array(self.problem.domains),
                )
                if not fix_top_choice_point(
                    self.domains_stk,
                    self.stks_top,
                    self.problem.variables_arr,
                    self.problem.offsets_arr,
                    variable_idx,
                    solution[variable_idx],
                    bound,
                ):
                    break
            else:
                logger.debug("Pruning choice points")
                if self.stks_top[0] == 0:
                    break
                self.stks_top[0] -= 1
                if not fix_choice_points(
                    self.domains_stk,
                    self.stks_top,
                    self.problem.variables_arr,
                    self.problem.offsets_arr,
                    variable_idx,
                    solution[variable_idx],
                    bound,
                ):
                    break
            reset_triggered_propagators(self.triggered_propagators, self.problem.propagator_nb)
        solution_queue.put((processor_idx, None, self.statistics))

    def solve_and_queue(self, processor_idx: int, solution_queue: Queue) -> None:
        """
        Enqueues the solutions.
        :param processor_idx: the index of the processor
        :param solution_queue: the solution queue
        """
        logger.info("Solving and queuing solutions found")
        compute_domains_addrs, var_heuristic_addrs, dom_heuristic_addrs, consistency_alg_addrs = (
            get_function_addresses()
        )
        while True:
            solution = solve_one(
                self.statistics,
                self.problem.algorithms,
                self.problem.bounds,
                self.problem.variables_arr,
                self.problem.offsets_arr,
                self.problem.props_variables,
                self.problem.props_offsets,
                self.problem.props_parameters,
                self.problem.triggers,
                self.domains_stk,
                self.not_entailed_propagators_stk,
                self.dom_update_stk,
                self.stks_top,
                self.triggered_propagators,
                self.consistency_alg_idx,
                self.decision_domains,
                self.var_heuristic_idx,
                self.var_heuristic_params,
                self.dom_heuristic_idx,
                self.dom_heuristic_params,
                compute_domains_addrs,
                consistency_alg_addrs,
                var_heuristic_addrs,
                dom_heuristic_addrs,
            )
            if solution is None:
                break
            solution_queue.put((processor_idx, solution, self.statistics))
            if not backtrack(
                self.statistics,
                self.not_entailed_propagators_stk,
                self.dom_update_stk,
                self.stks_top,
                self.triggered_propagators,
                self.problem.triggers,
            ):
                break
        solution_queue.put((processor_idx, None, self.statistics))


@njit(cache=True)
def solve_one(
    statistics: NDArray,
    algorithms: NDArray,
    bounds: NDArray,
    variables_arr: NDArray,
    offsets_arr: NDArray,
    props_variables: NDArray,
    props_offsets: NDArray,
    props_parameters: NDArray,
    triggers: NDArray,
    domains_stk: NDArray,
    not_entailed_propagators_stk: NDArray,
    dom_update_stk: NDArray,
    stks_top: NDArray,
    triggered_propagators: NDArray,
    consistency_alg_idx: int,
    decision_domains: NDArray,
    var_heuristic_idx: int,
    var_heuristic_params: NDArray,
    dom_heuristic_idx: int,
    dom_heuristic_params: NDArray,
    compute_domains_addrs: NDArray,
    consistency_alg_addrs: NDArray,
    var_heuristic_addrs: NDArray,
    dom_heuristic_addrs: NDArray,
) -> Optional[NDArray]:
    """
    Find at most one solution.
    :param statistics: a Numpy array of statistics
    :param algorithms: the algorithms indexed by propagators
    :param bounds: the bounds indexed by propagators
    :param variables_arr: the domain indices indexed by variables
    :param offsets_arr: the domain offsets indexed by variables
    :param props_variables: the domain indices indexed by propagator variables
    :param props_offsets: the domain offsets indexed by propagator variables
    :param props_parameters: the parameters indexed by propagator variables
    :param triggers: a Numpy array of event masks indexed by shared domain indices and propagators
    :param domains_stk: a stack of shared domains;
    the first level correspond to the current shared domains, the rest correspond to the choice points
    :param not_entailed_propagators_stk: a stack not entailed propagators;
    the first level correspond to the propagators currently not entailed, the rest correspond to the choice points
    :param dom_update_stk: the stack of domain updates
    :param stks_top: the index of the top of the stacks as a Numpy array
    :param triggered_propagators: the Numpy array of triggered propagators
    :param consistency_alg_idx: the index of the consistency algorithm
    :param decision_domains: the indices of the shared domains on which decisions will be made
    :param var_heuristic_idx: the index of the variable heuristic
    :param var_heuristic_params: a list of lists of parameters,
    usually parameters are costs and there is a list of value costs per variable/shared domain
    :param dom_heuristic_idx: the index of the domain heuristic
    :param dom_heuristic_params: a list of lists of parameters,
    usually parameters are costs and there is a list of value costs per variable/shared domain
    :param compute_domains_addrs: the addresses of the compute_domains functions
    :param consistency_alg_addrs: the addresses of the functions implementing the consistency algorithms
    :param var_heuristic_addrs: the addresses of the functions implementing the variable heuristics
    :param dom_heuristic_addrs: the addresses of the functions implementing the domain heuristics
    :return: the solution if it exists or None
    """
    if NUMBA_DISABLE_JIT:
        consistency_alg_fct = CONSISTENCY_ALG_FCTS[consistency_alg_idx]
        var_heuristic_fct = VAR_HEURISTIC_FCTS[var_heuristic_idx]
        dom_heuristic_fct = DOM_HEURISTIC_FCTS[dom_heuristic_idx]
    else:
        consistency_alg_fct = function_from_address(TYPE_CONSISTENCY_ALG, consistency_alg_addrs[consistency_alg_idx])
        var_heuristic_fct = function_from_address(TYPE_VAR_HEURISTIC, var_heuristic_addrs[var_heuristic_idx])
        dom_heuristic_fct = function_from_address(TYPE_DOM_HEURISTIC, dom_heuristic_addrs[dom_heuristic_idx])
    while True:
        status = consistency_alg_fct(
            statistics,
            algorithms,
            bounds,
            variables_arr,
            offsets_arr,
            props_variables,
            props_offsets,
            props_parameters,
            triggers,
            domains_stk,
            not_entailed_propagators_stk,
            dom_update_stk,
            stks_top,
            triggered_propagators,
            compute_domains_addrs,
            decision_domains,
        )
        if status == PROBLEM_BOUND:
            statistics[STATS_IDX_SOLUTION_NB] += 1
            return get_solution(domains_stk, stks_top, variables_arr, offsets_arr)
        elif status == PROBLEM_UNBOUND:
            dom_idx = var_heuristic_fct(decision_domains, domains_stk, stks_top, var_heuristic_params)
            events = dom_heuristic_fct(
                domains_stk,
                not_entailed_propagators_stk,
                dom_update_stk,
                stks_top,
                dom_idx,
                dom_heuristic_params,
            )
            update_propagators(
                triggered_propagators, not_entailed_propagators_stk[stks_top[0]], triggers, events, dom_idx
            )
            statistics[STATS_IDX_SOLVER_CHOICE_NB] += 1
            if stks_top[0] > statistics[STATS_IDX_SOLVER_CHOICE_DEPTH]:
                statistics[STATS_IDX_SOLVER_CHOICE_DEPTH] = stks_top[0]
        elif not backtrack(
            statistics,
            not_entailed_propagators_stk,
            dom_update_stk,
            stks_top,
            triggered_propagators,
            triggers,
        ):
            return None


def get_function_addresses() -> Tuple[NDArray, NDArray, NDArray, NDArray]:
    """
    Returns the addresses of the
    compute_domains, variable heuristics, domain heuristics and consistency algorithms functions.
    :return: a tuple of NDArrays
    """
    if NUMBA_DISABLE_JIT:
        return np.empty(0), np.empty(0), np.empty(0), np.empty(0)
    return (
        np.array(build_function_address_list(COMPUTE_DOMAINS_FCTS, SIGNATURE_COMPUTE_DOMAINS)),
        np.array(build_function_address_list(VAR_HEURISTIC_FCTS, SIGNATURE_VAR_HEURISTIC)),
        np.array(build_function_address_list(DOM_HEURISTIC_FCTS, SIGNATURE_DOM_HEURISTIC)),
        np.array(build_function_address_list(CONSISTENCY_ALG_FCTS, SIGNATURE_CONSISTENCY_ALG)),
    )



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from typing import Callable, List, Optional, Tuple, Union

import numpy as np


class PropagatorTest:
    def assert_compute_domains(
        self,
        compute_domains_fct: Callable,
        domains: List[Union[int, Tuple[int, int]]],
        parameters: List[int],
        consistency_result: int,
        expected_domains: Optional[List[List[int]]],
    ) -> None:
        domains_arr = np.array(
            [(domain, domain) if isinstance(domain, int) else domain for domain in domains], dtype=np.int32
        )
        assert compute_domains_fct(domains_arr, np.array(parameters, dtype=np.int32)) == consistency_result
        if expected_domains:
            assert np.all(domains_arr == np.array(expected_domains))



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import argparse

from nucs.examples.default_argument_parser import DefaultArgumentParser
from nucs.examples.queens.queens_problem import QueensProblem
from nucs.heuristics.heuristics import (
    DOM_HEURISTIC_MID_VALUE,
    VAR_HEURISTIC_FIRST_NOT_INSTANTIATED,
    VAR_HEURISTIC_SMALLEST_DOMAIN,
)
from nucs.solvers.backtrack_solver import BacktrackSolver
from nucs.solvers.consistency_algorithms import CONSISTENCY_ALG_BC, CONSISTENCY_ALG_SHAVING
from nucs.solvers.multiprocessing_solver import MultiprocessingSolver

# Run with the following command (the second run is much faster because the code has been compiled):
# NUMBA_CACHE_DIR=.numba/cache python -m nucs.examples.queens -n 10
if __name__ == "__main__":
    parser = DefaultArgumentParser()
    parser.add_argument("--ff", type=bool, action=argparse.BooleanOptionalAction, default=False)
    parser.add_argument("-n", type=int, default=10)
    args = parser.parse_args()
    problem = QueensProblem(args.n)
    solver = (
        MultiprocessingSolver(
            [
                BacktrackSolver(
                    problem,
                    consistency_alg_idx=CONSISTENCY_ALG_SHAVING if args.shaving else CONSISTENCY_ALG_BC,
                    var_heuristic_idx=(
                        VAR_HEURISTIC_SMALLEST_DOMAIN if args.ff else VAR_HEURISTIC_FIRST_NOT_INSTANTIATED
                    ),
                    log_level=args.log_level,
                )
                for problem in problem.split(args.processors, 0)
            ]
        )
        if args.processors > 1
        else BacktrackSolver(
            problem,
            consistency_alg_idx=CONSISTENCY_ALG_SHAVING if args.shaving else CONSISTENCY_ALG_BC,
            var_heuristic_idx=VAR_HEURISTIC_SMALLEST_DOMAIN if args.ff else VAR_HEURISTIC_FIRST_NOT_INSTANTIATED,
            dom_heuristic_idx=DOM_HEURISTIC_MID_VALUE,
            log_level=args.log_level,
            stks_max_height=2048,
        )
    )
    if args.all:
        solver.solve_all()
        solver.print_statistics()
    else:
        solution = solver.find_one()
        if args.stats:
            solver.print_statistics()
        if args.display:
            problem.print_solution(solution)



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from typing import Any

from numpy.typing import NDArray

from nucs.problems.problem import Problem
from nucs.propagators.propagators import ALG_ABS_EQ, ALG_AFFINE_EQ, ALG_AFFINE_LEQ, ALG_ALLDIFFERENT


class AllIntervalSeriesProblem(Problem):
    """
    CSPLIB problem #7 - https://www.csplib.org/Problems/prob007/
    """

    def __init__(self, n: int, symmetry_breaking: bool):
        """
        Initializes the problem.
        :param n: the size of the sequence
        :param symmetry_breaking: a boolean indicating if symmetry constraints should be added to the model
        """
        super().__init__([(0, n - 1)] * n + [(-n + 1, n - 1)] * (n - 1) + [(1, n - 1)] * (n - 1))
        self.n = n
        for i in range(n - 1):
            self.add_propagator(([n + i, i + 1, i], ALG_AFFINE_EQ, [1, -1, 1, 0]))
            self.add_propagator(([n + i, 2 * n - 1 + i], ALG_ABS_EQ, []))
        self.add_propagator((list(range(n)), ALG_ALLDIFFERENT, []))
        self.add_propagator((list(range(2 * n - 1, 3 * n - 2)), ALG_ALLDIFFERENT, []))
        if symmetry_breaking:
            self.add_propagator(([0, 1], ALG_AFFINE_LEQ, [1, -1, -1]))
            self.add_propagator(([3 * n - 3, 2 * n - 1], ALG_AFFINE_LEQ, [1, -1, -1]))

    def solution_as_printable(self, solution: NDArray) -> Any:
        return solution[: self.n].tolist()



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from typing import List

import pytest

from nucs.constants import STATS_IDX_SOLUTION_NB
from nucs.examples.all_interval_series.all_interval_series_problem import AllIntervalSeriesProblem
from nucs.solvers.backtrack_solver import BacktrackSolver


class TestAllIntervalSeries:
    @pytest.mark.parametrize(
        "size,values",
        [
            (8, [0, 7, 1, 6, 2, 5, 3, 4]),
            (9, [0, 8, 1, 7, 2, 6, 3, 5, 4]),
        ],
    )
    def test_find_one(self, size: int, values: List[int]) -> None:
        problem = AllIntervalSeriesProblem(size, True)
        solver = BacktrackSolver(problem)
        solution = solver.find_one()
        assert solution is not None
        assert solution[:size].tolist() == values

    @pytest.mark.parametrize(
        "size,solution_nb",
        [
            (3, 1),
            (4, 1),
            (5, 2),
            (6, 6),
            (7, 8),
            (8, 10),
            (9, 30),
        ],
    )
    def test_solve_all(self, size: int, solution_nb: int) -> None:
        problem = AllIntervalSeriesProblem(size, True)
        solver = BacktrackSolver(problem)
        solver.solve_all()
        assert solver.statistics[STATS_IDX_SOLUTION_NB] == solution_nb



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import EVENT_MASK_MAX, EVENT_MASK_MIN_MAX, MAX, MIN, PROP_CONSISTENCY, PROP_INCONSISTENCY


def get_complexity_abs_eq(n: int, parameters: NDArray) -> float:
    """
    Returns the time complexity of the propagator as a float.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: a float
    """
    return 1


@njit(cache=True)
def get_triggers_abs_eq(n: int, dom_idx: int, parameters: NDArray) -> int:
    """
    Returns the triggers for this propagator.
    :param parameters: the parameters, unused here
    :return: an array of triggers
    """
    if dom_idx == 0:
        return EVENT_MASK_MIN_MAX
    return EVENT_MASK_MAX


@njit(cache=True)
def compute_domains_abs_eq(domains: NDArray, parameters: NDArray) -> int:
    """
    Implements abs(y)=x.
    :param domains: the domains of the variables, y is the first domain, x the second
    :param parameters: unused here
    :return: the status of the propagation (consistency, inconsistency or entailment) as an int
    """
    y = domains[0]
    x = domains[1]
    if y[MIN] > 0:
        x[MIN] = max(x[MIN], y[MIN])
        x[MAX] = min(x[MAX], y[MAX])
        if x[MIN] > x[MAX]:
            return PROP_INCONSISTENCY
        y[MIN] = max(y[MIN], x[MIN])
        y[MAX] = min(y[MAX], x[MAX])
        if y[MIN] > y[MAX]:
            return PROP_INCONSISTENCY
    elif y[MAX] < 0:
        x[MIN] = max(x[MIN], -y[MAX])
        x[MAX] = min(x[MAX], -y[MIN])
        if x[MIN] > x[MAX]:
            return PROP_INCONSISTENCY
        y[MIN] = max(y[MIN], -x[MAX])
        y[MAX] = min(y[MAX], -x[MIN])
        if y[MIN] > y[MAX]:
            return PROP_INCONSISTENCY
    else:
        x[MIN] = max(x[MIN], 0)
        x[MAX] = min(x[MAX], max(-y[MIN], y[MAX]))
        if x[MIN] > x[MAX]:
            return PROP_INCONSISTENCY
        y[MIN] = max(y[MIN], -x[MAX])
        y[MAX] = min(y[MAX], x[MAX])
        if y[MIN] > y[MAX]:
            return PROP_INCONSISTENCY
    return PROP_CONSISTENCY



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################

from nucs.constants import OPTIM_MODES, OPTIM_RESET
from nucs.examples.default_argument_parser import DefaultArgumentParser
from nucs.examples.knapsack.knapsack_problem import KnapsackProblem
from nucs.heuristics.heuristics import DOM_HEURISTIC_MAX_VALUE, VAR_HEURISTIC_FIRST_NOT_INSTANTIATED
from nucs.solvers.backtrack_solver import BacktrackSolver

# Run with the following command (the second run is much faster because the code has been compiled):
# NUMBA_CACHE_DIR=.numba/cache python -m nucs.examples.knapsack
if __name__ == "__main__":
    parser = DefaultArgumentParser()
    parser.add_argument("--opt_mode", choices=OPTIM_MODES, default=OPTIM_RESET)
    args = parser.parse_args()
    problem = KnapsackProblem(
        [40, 40, 38, 38, 36, 36, 34, 34, 32, 32, 30, 30, 28, 28, 26, 26, 24, 24, 22, 22],
        [40, 40, 38, 38, 36, 36, 34, 34, 32, 32, 30, 30, 28, 28, 26, 26, 24, 24, 22, 22],
        55,
    )
    solver = BacktrackSolver(
        problem,
        var_heuristic_idx=VAR_HEURISTIC_FIRST_NOT_INSTANTIATED,
        dom_heuristic_idx=DOM_HEURISTIC_MAX_VALUE,
        log_level=args.log_level,
    )
    solution = solver.maximize(problem.weight, mode=args.opt_mode)
    if args.stats:
        solver.print_statistics()
    if args.display:
        problem.print_solution(solution)



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import (
    DOM_UPDATE_EVENTS,
    DOM_UPDATE_IDX,
    EVENT_MASK_MAX,
    EVENT_MASK_MAX_GROUND,
    EVENT_MASK_MIN,
    EVENT_MASK_MIN_GROUND,
    EVENT_MASK_MIN_MAX_GROUND,
    MAX,
    MIN,
)
from nucs.heuristics.max_value_dom_heuristic import max_value_dom_heuristic
from nucs.heuristics.min_value_dom_heuristic import min_value_dom_heuristic
from nucs.solvers.choice_points import cp_put


@njit(cache=True)
def value_dom_heuristic(
    domains_stk: NDArray,
    not_entailed_propagators_stk: NDArray,
    dom_update_stk: NDArray,
    stks_top: NDArray,
    dom_idx: int,
    value: int,
    params: NDArray,
) -> int:
    """
    Chooses a value given as a parameter.
    :param domains_stk: the stack of shared domains
    :param not_entailed_propagators_stk: the stack of not entailed propagators
    :param dom_update_stk: the stack of domain updates
    :param stks_top: the index of the top of the stacks as a Numpy array
    :param dom_idx: the index of the shared domain
    :param value: the value
    :param params: a two-dimensional parameter array
    :return: the events
    """
    top = stks_top[0]
    if value == domains_stk[top, dom_idx, MIN]:
        return min_value_dom_heuristic(
            domains_stk, not_entailed_propagators_stk, dom_update_stk, stks_top, dom_idx, params
        )
    if value == domains_stk[top, dom_idx, MAX]:
        return max_value_dom_heuristic(
            domains_stk, not_entailed_propagators_stk, dom_update_stk, stks_top, dom_idx, params
        )
    cp_put(domains_stk, not_entailed_propagators_stk, stks_top)
    cp_put(domains_stk, not_entailed_propagators_stk, stks_top)
    domains_stk[top + 2, dom_idx, :] = value
    domains_stk[top + 1, dom_idx, MAX] = value - 1
    domains_stk[top, dom_idx, MIN] = value + 1
    dom_update_stk[top + 1, DOM_UPDATE_IDX] = dom_update_stk[top, DOM_UPDATE_IDX] = dom_idx
    dom_update_stk[top + 1, DOM_UPDATE_EVENTS] = (
        EVENT_MASK_MAX_GROUND
        if domains_stk[top + 1, dom_idx, MIN] == domains_stk[top + 1, dom_idx, MAX]
        else EVENT_MASK_MAX
    )
    dom_update_stk[top, DOM_UPDATE_EVENTS] = (
        EVENT_MASK_MIN_GROUND if domains_stk[top, dom_idx, MIN] == domains_stk[top, dom_idx, MAX] else EVENT_MASK_MIN
    )
    return EVENT_MASK_MIN_MAX_GROUND



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import argparse

from nucs.constants import LOG_LEVEL_INFO, LOG_LEVELS


class DefaultArgumentParser(argparse.ArgumentParser):
    def __init__(self) -> None:
        super().__init__()
        self.add_argument(
            "--all", help="find all solutions", type=bool, action=argparse.BooleanOptionalAction, default=True
        )
        self.add_argument(
            "--log_level",
            help="set the log level",
            choices=LOG_LEVELS,
            default=LOG_LEVEL_INFO,
        )
        self.add_argument(
            "--processors",
            help="set the number of processors",
            type=int,
            default=1,
        )
        self.add_argument(
            "--display",
            help="display the solution(s)",
            type=bool,
            action=argparse.BooleanOptionalAction,
            default=True,
        )
        self.add_argument(
            "--stats",
            help="display the statistics",
            type=bool,
            action=argparse.BooleanOptionalAction,
            default=True,
        )
        self.add_argument(
            "--shaving",
            help="use shaving together with bound consistency",
            type=bool,
            action=argparse.BooleanOptionalAction,
            default=False,
        )
        self.add_argument(
            "--symmetry_breaking",
            help="add symmetry breaking constraints",
            action=argparse.BooleanOptionalAction,
            default=True,
        )



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from typing import List, Optional, Tuple, Union

import numpy as np
import pytest

from nucs.constants import PROP_CONSISTENCY
from nucs.propagators.alldifferent_propagator import compute_domains_alldifferent, path_max, path_min, path_set
from tests.propagators.propagator_test import PropagatorTest


class TestAlldifferent(PropagatorTest):
    def test_path_min(self) -> None:
        a = np.array([2, 3, 4, 0, 1])
        assert path_min(a, 0) == 0
        assert path_min(a, 1) == 1
        assert path_min(a, 2) == 2
        assert path_min(a, 3) == 0
        assert path_min(a, 4) == 1

    def test_path_max(self) -> None:
        a = np.array([2, 3, 4, 0, 1])
        assert path_max(a, 0) == 4
        assert path_max(a, 1) == 3
        assert path_max(a, 2) == 4
        assert path_max(a, 3) == 3
        assert path_max(a, 4) == 4

    def test_path_set(self) -> None:
        a = np.array([2, 3, 4, 0, 1])
        path_set(a, 0, 4, -1)
        assert np.all(a == np.array([-1, 3, -1, 0, 1]))

    @pytest.mark.parametrize(
        "domains,parameters,consistency_result,expected_domains",
        [
            (
                [(3, 6), (3, 4), (2, 5), (2, 4), (3, 4), (1, 6)],
                [],
                PROP_CONSISTENCY,
                [[6, 6], [3, 4], [5, 5], [2, 2], [3, 4], [1, 1]],
            ),
            ([(0, 0), (2, 2), (1, 2)], [], PROP_CONSISTENCY, [[0, 0], [2, 2], [1, 1]]),
            ([(0, 0), (0, 4), (0, 4), (0, 4), (0, 4)], [], PROP_CONSISTENCY, [[0, 0], [1, 4], [1, 4], [1, 4], [1, 4]]),
        ],
    )
    def test_compute_domains(
        self,
        domains: List[Union[int, Tuple[int, int]]],
        parameters: List[int],
        consistency_result: int,
        expected_domains: Optional[List[List[int]]],
    ) -> None:
        self.assert_compute_domains(
            compute_domains_alldifferent, domains, parameters, consistency_result, expected_domains
        )



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from typing import Callable

from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.heaps import min_heap_add
from nucs.propagators.abs_eq_propagator import compute_domains_abs_eq, get_complexity_abs_eq, get_triggers_abs_eq
from nucs.propagators.affine_eq_propagator import (
    compute_domains_affine_eq,
    get_complexity_affine_eq,
    get_triggers_affine_eq,
)
from nucs.propagators.affine_geq_propagator import (
    compute_domains_affine_geq,
    get_complexity_affine_geq,
    get_triggers_affine_geq,
)
from nucs.propagators.affine_leq_propagator import (
    compute_domains_affine_leq,
    get_complexity_affine_leq,
    get_triggers_affine_leq,
)
from nucs.propagators.alldifferent_propagator import (
    compute_domains_alldifferent,
    get_complexity_alldifferent,
    get_triggers_alldifferent,
)
from nucs.propagators.and_eq_propagator import compute_domains_and_eq, get_complexity_and_eq, get_triggers_and_eq
from nucs.propagators.count_eq_c_propagator import (
    compute_domains_count_eq_c,
    get_complexity_count_eq_c,
    get_triggers_count_eq_c,
)
from nucs.propagators.count_eq_propagator import (
    compute_domains_count_eq,
    get_complexity_count_eq,
    get_triggers_count_eq,
)
from nucs.propagators.count_geq_c_propagator import (
    compute_domains_count_geq_c,
    get_complexity_count_geq_c,
    get_triggers_count_geq_c,
)
from nucs.propagators.count_leq_c_propagator import (
    compute_domains_count_leq_c,
    get_complexity_count_leq_c,
    get_triggers_count_leq_c,
)
from nucs.propagators.dummy_propagator import compute_domains_dummy, get_complexity_dummy, get_triggers_dummy
from nucs.propagators.element_eq_propagator import (
    compute_domains_element_eq,
    get_complexity_element_eq,
    get_triggers_element_eq,
)
from nucs.propagators.element_l_eq_alldifferent_propagator import (
    compute_domains_element_l_eq_alldifferent,
    get_complexity_element_l_eq_alldifferent,
    get_triggers_element_l_eq_alldifferent,
)
from nucs.propagators.element_l_eq_c_alldifferent_propagator import (
    compute_domains_element_l_eq_c_alldifferent,
    get_complexity_element_l_eq_c_alldifferent,
    get_triggers_element_l_eq_c_alldifferent,
)
from nucs.propagators.element_l_eq_c_propagator import (
    compute_domains_element_l_eq_c,
    get_complexity_element_l_eq_c,
    get_triggers_element_l_eq_c,
)
from nucs.propagators.element_l_eq_propagator import (
    compute_domains_element_l_eq,
    get_complexity_element_l_eq,
    get_triggers_element_l_eq,
)
from nucs.propagators.equiv_eq_propagator import (
    compute_domains_equiv_eq,
    get_complexity_equiv_eq,
    get_triggers_equiv_eq,
)
from nucs.propagators.gcc_propagator import compute_domains_gcc, get_complexity_gcc, get_triggers_gcc
from nucs.propagators.lexicographic_leq_propagator import (
    compute_domains_lexicographic_leq,
    get_complexity_lexicographic_leq,
    get_triggers_lexicographic_leq,
)
from nucs.propagators.max_eq_propagator import compute_domains_max_eq, get_complexity_max_eq, get_triggers_max_eq
from nucs.propagators.max_leq_propagator import compute_domains_max_leq, get_complexity_max_leq, get_triggers_max_leq
from nucs.propagators.min_eq_propagator import compute_domains_min_eq, get_complexity_min_eq, get_triggers_min_eq
from nucs.propagators.min_geq_propagator import compute_domains_min_geq, get_complexity_min_geq, get_triggers_min_geq
from nucs.propagators.no_sub_cycle_propagator import (
    compute_domains_no_sub_cycle,
    get_complexity_no_sub_cycle,
    get_triggers_no_sub_cycle,
)
from nucs.propagators.permutation_aux_propagator import (
    compute_domains_permutation_aux,
    get_complexity_permutation_aux,
    get_triggers_permutation_aux,
)
from nucs.propagators.relation_propagator import (
    compute_domains_relation,
    get_complexity_relation,
    get_triggers_relation,
)
from nucs.propagators.scc_propagator import compute_domains_scc, get_complexity_scc, get_triggers_scc

GET_TRIGGERS_FCTS = []
GET_COMPLEXITY_FCTS = []
COMPUTE_DOMAINS_FCTS = []


def register_propagator(get_triggers_fct: Callable, get_complexity_fct: Callable, compute_domains_fct: Callable) -> int:
    """
    Registers a propagator by adding its 3 functions to the corresponding lists of functions.
    :param get_triggers_fct: a function that returns the triggers
    :param get_complexity_fct: a function that computes the complexity
    :param compute_domains_fct: a function that computes the domains
    :return: the index of the propagator
    """
    GET_TRIGGERS_FCTS.append(get_triggers_fct)
    GET_COMPLEXITY_FCTS.append(get_complexity_fct)
    COMPUTE_DOMAINS_FCTS.append(compute_domains_fct)
    return len(COMPUTE_DOMAINS_FCTS) - 1


ALG_ABS_EQ = register_propagator(get_triggers_abs_eq, get_complexity_abs_eq, compute_domains_abs_eq)
ALG_AND_EQ = register_propagator(get_triggers_and_eq, get_complexity_and_eq, compute_domains_and_eq)
ALG_AFFINE_EQ = register_propagator(get_triggers_affine_eq, get_complexity_affine_eq, compute_domains_affine_eq)
ALG_AFFINE_GEQ = register_propagator(get_triggers_affine_geq, get_complexity_affine_geq, compute_domains_affine_geq)
ALG_AFFINE_LEQ = register_propagator(get_triggers_affine_leq, get_complexity_affine_leq, compute_domains_affine_leq)
ALG_ALLDIFFERENT = register_propagator(
    get_triggers_alldifferent, get_complexity_alldifferent, compute_domains_alldifferent
)
ALG_COUNT_EQ = register_propagator(get_triggers_count_eq, get_complexity_count_eq, compute_domains_count_eq)
ALG_COUNT_EQ_C = register_propagator(get_triggers_count_eq_c, get_complexity_count_eq_c, compute_domains_count_eq_c)
ALG_COUNT_GEQ_C = register_propagator(get_triggers_count_geq_c, get_complexity_count_geq_c, compute_domains_count_geq_c)
ALG_COUNT_LEQ_C = register_propagator(get_triggers_count_leq_c, get_complexity_count_leq_c, compute_domains_count_leq_c)
ALG_DUMMY = register_propagator(get_triggers_dummy, get_complexity_dummy, compute_domains_dummy)
ALG_ELEMENT_EQ = register_propagator(get_triggers_element_eq, get_complexity_element_eq, compute_domains_element_eq)
ALG_ELEMENT_L_EQ = register_propagator(
    get_triggers_element_l_eq, get_complexity_element_l_eq, compute_domains_element_l_eq
)
ALG_ELEMENT_L_EQ_ALLDIFFERENT = register_propagator(
    get_triggers_element_l_eq_alldifferent,
    get_complexity_element_l_eq_alldifferent,
    compute_domains_element_l_eq_alldifferent,
)
ALG_ELEMENT_L_EQ_C = register_propagator(
    get_triggers_element_l_eq_c, get_complexity_element_l_eq_c, compute_domains_element_l_eq_c
)
ALG_ELEMENT_L_EQ_C_ALLDIFFERENT = register_propagator(
    get_triggers_element_l_eq_c_alldifferent,
    get_complexity_element_l_eq_c_alldifferent,
    compute_domains_element_l_eq_c_alldifferent,
)
ALG_EQUIV_EQ = register_propagator(get_triggers_equiv_eq, get_complexity_equiv_eq, compute_domains_equiv_eq)
ALG_GCC = register_propagator(get_triggers_gcc, get_complexity_gcc, compute_domains_gcc)
ALG_LEXICOGRAPHIC_LEQ = register_propagator(
    get_triggers_lexicographic_leq, get_complexity_lexicographic_leq, compute_domains_lexicographic_leq
)
ALG_MAX_EQ = register_propagator(get_triggers_max_eq, get_complexity_max_eq, compute_domains_max_eq)
ALG_MAX_LEQ = register_propagator(get_triggers_max_leq, get_complexity_max_leq, compute_domains_max_leq)
ALG_MIN_EQ = register_propagator(get_triggers_min_eq, get_complexity_min_eq, compute_domains_min_eq)
ALG_MIN_GEQ = register_propagator(get_triggers_min_geq, get_complexity_min_geq, compute_domains_min_geq)
ALG_NO_SUB_CYCLE = register_propagator(
    get_triggers_no_sub_cycle, get_complexity_no_sub_cycle, compute_domains_no_sub_cycle
)
ALG_PERMUTATION_AUX = register_propagator(
    get_triggers_permutation_aux, get_complexity_permutation_aux, compute_domains_permutation_aux
)
ALG_RELATION = register_propagator(get_triggers_relation, get_complexity_relation, compute_domains_relation)
ALG_SCC = register_propagator(get_triggers_scc, get_complexity_scc, compute_domains_scc)


@njit(cache=True)
def reset_triggered_propagators(triggered_propagators: NDArray, propagator_nb: int) -> None:
    triggered_propagators[:] = 0
    for prop_idx in range(propagator_nb):
        min_heap_add(triggered_propagators, propagator_nb, prop_idx)


@njit(cache=True)
def update_propagators(
    triggered_propagators: NDArray,
    not_entailed_propagators: NDArray,
    triggers: NDArray,
    events: int,
    dom_idx: int,
    previous_prop_idx: int = -1,
) -> None:
    propagator_nb = len(not_entailed_propagators)
    for prop_idx in triggers[dom_idx, events]:
        if prop_idx == -1:
            break
        if not_entailed_propagators[prop_idx] and prop_idx != previous_prop_idx:
            min_heap_add(triggered_propagators, propagator_nb, prop_idx)



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import (
    EVENT_MASK_GROUND,
    MAX,
    MIN,
    PROBLEM_INCONSISTENT,
    PROBLEM_UNBOUND,
    STATS_IDX_ALG_BC_WITH_SHAVING_NB,
    STATS_IDX_ALG_SHAVING_CHANGE_NB,
    STATS_IDX_ALG_SHAVING_NB,
    STATS_IDX_ALG_SHAVING_NO_CHANGE_NB,
)
from nucs.heuristics.first_not_instantiated_var_heuristic import first_not_instantiated_var_heuristic
from nucs.heuristics.heuristics import max_value_dom_heuristic, min_value_dom_heuristic
from nucs.propagators.propagators import update_propagators
from nucs.solvers.bound_consistency_algorithm import bound_consistency_algorithm
from nucs.solvers.choice_points import backtrack


@njit(cache=True)
def shave_bound(
    bound: int,
    dom_idx: int,
    statistics: NDArray,
    algorithms: NDArray,
    bounds: NDArray,
    variables_arr: NDArray,
    offsets_arr: NDArray,
    props_variables: NDArray,
    props_offsets: NDArray,
    props_parameters: NDArray,
    triggers: NDArray,
    domains_stk: NDArray,
    not_entailed_propagators_stk: NDArray,
    dom_update_stk: NDArray,
    stks_top: NDArray,
    triggered_propagators: NDArray,
    compute_domains_addrs: NDArray,
    decision_variables: NDArray,
) -> bool:
    events = (
        max_value_dom_heuristic(domains_stk, not_entailed_propagators_stk, dom_update_stk, stks_top, dom_idx, None)
        if bound == MAX
        else min_value_dom_heuristic(domains_stk, not_entailed_propagators_stk, dom_update_stk, stks_top, dom_idx, None)
    )
    if domains_stk[stks_top[0], dom_idx, MIN] == domains_stk[stks_top[0], dom_idx, MAX]:
        events |= EVENT_MASK_GROUND
    update_propagators(triggered_propagators, not_entailed_propagators_stk[stks_top[0]], triggers, events, dom_idx)
    if (
        bound_consistency_algorithm(
            statistics,
            algorithms,
            bounds,
            variables_arr,
            offsets_arr,
            props_variables,
            props_offsets,
            props_parameters,
            triggers,
            domains_stk,
            not_entailed_propagators_stk,
            dom_update_stk,
            stks_top,
            triggered_propagators,
            compute_domains_addrs,
            decision_variables,
        )
        == PROBLEM_INCONSISTENT
    ):
        has_shaved = True
    else:
        domains_stk[stks_top[0] - 1, dom_idx, bound] += 1 if bound == MAX else -1
        has_shaved = False
    backtrack(
        statistics,
        not_entailed_propagators_stk,
        dom_update_stk,
        stks_top,
        triggered_propagators,
        triggers,
    )
    return has_shaved


@njit(cache=True)
def shaving_consistency_algorithm(
    statistics: NDArray,
    algorithms: NDArray,
    bounds: NDArray,
    variables_arr: NDArray,
    offsets_arr: NDArray,
    props_variables: NDArray,
    props_offsets: NDArray,
    props_parameters: NDArray,
    triggers: NDArray,
    domains_stk: NDArray,
    not_entailed_propagators_stk: NDArray,
    dom_update_stk: NDArray,
    stks_top: NDArray,
    triggered_propagators: NDArray,
    compute_domains_addrs: NDArray,
    decision_domains: NDArray,
) -> int:
    """
    Shaving consistency algorithm.
    :param statistics: a Numpy array of statistics
    :param algorithms: the algorithms indexed by propagators
    :param bounds: the bounds indexed by propagators
    :param variables_arr: the domain indices indexed by variables
    :param offsets_arr: the domain offsets indexed by variables
    :param props_variables: the domain indices indexed by propagator variables
    :param props_offsets: the domain offsets indexed by propagator variables
    :param props_parameters: the parameters indexed by propagator variables
    :param triggers: a Numpy array of event masks indexed by shared domain indices and propagators
    :param domains_stk: a stack of shared domains;
    the first level correspond to the current shared domains, the rest correspond to the choice points
    :param not_entailed_propagators_stk: a stack not entailed propagators;
    the first level correspond to the propagators currently not entailed, the rest correspond to the choice points
    :param dom_update_stk: the stack of domain updates
    :param stks_top: the height of the stacks as a Numpy array
    :param triggered_propagators: the Numpy array of triggered propagators
    :param compute_domains_addrs: the addresses of the compute_domains functions
    :param decision_domains: the indices of the shared domains on which decisions will be made
    :return: a status (consistency, inconsistency or entailment) as an integer
    """
    statistics[STATS_IDX_ALG_BC_WITH_SHAVING_NB] += 1
    shr_domains_nb = len(domains_stk[0])
    bound = MIN
    has_shaved = True
    start_idx = 0
    while start_idx < shr_domains_nb:
        if has_shaved:
            status = bound_consistency_algorithm(
                statistics,
                algorithms,
                bounds,
                variables_arr,
                offsets_arr,
                props_variables,
                props_offsets,
                props_parameters,
                triggers,
                domains_stk,
                not_entailed_propagators_stk,
                dom_update_stk,
                stks_top,
                triggered_propagators,
                compute_domains_addrs,
                decision_domains,
            )
            if status != PROBLEM_UNBOUND:
                return status
        dom_idx = first_not_instantiated_var_heuristic(
            decision_domains[decision_domains >= start_idx], domains_stk, stks_top, None
        )
        if dom_idx == -1:  # all variables after start_idx are instantiated
            break
        statistics[STATS_IDX_ALG_SHAVING_NB] += 1
        has_shaved = shave_bound(
            bound,
            dom_idx,
            statistics,
            algorithms,
            bounds,
            variables_arr,
            offsets_arr,
            props_variables,
            props_offsets,
            props_parameters,
            triggers,
            domains_stk,
            not_entailed_propagators_stk,
            dom_update_stk,
            stks_top,
            triggered_propagators,
            compute_domains_addrs,
            decision_domains,
        )
        start_idx = dom_idx
        if has_shaved:
            statistics[STATS_IDX_ALG_SHAVING_CHANGE_NB] += 1
        else:
            statistics[STATS_IDX_ALG_SHAVING_NO_CHANGE_NB] += 1
            # this is one of the many shaving strategies
            bound += 1
            if bound > MAX:
                bound = MIN
                start_idx += 1
    return PROBLEM_UNBOUND



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from typing import Any

from numpy.typing import NDArray

from nucs.problems.problem import Problem
from nucs.propagators.propagators import ALG_ALLDIFFERENT, ALG_PERMUTATION_AUX


class QueensProblem(Problem):
    """
    A simple model for the n-queens problem.

    CSPLIB problem #54 - https://www.csplib.org/Problems/prob054/
    """

    def __init__(self, n: int):
        """
        Initializes the problem.
        :param n: the number of queens
        """
        super().__init__([(0, n - 1)] * n, list(range(n)) * 3, [0] * n + list(range(n)) + list(range(0, -n, -1)))
        self.n = n
        self.add_propagators([(list(range(i * n, i * n + n)), ALG_ALLDIFFERENT, []) for i in range(3)])

    def solution_as_printable(self, solution: NDArray) -> Any:
        return [([" "] * i + ["Q"] + [" "] * (self.n - i - 1)) for i in (solution[: self.n])]


class QueensDualProblem(Problem):
    """
    A dual model for the n-queens problem.

    CSPLIB problem #54 - https://www.csplib.org/Problems/prob054/
    """

    def __init__(self, n: int):
        """
        Initializes the problem.
        :param n: the number of queens
        """
        super().__init__(
            [(0, n - 1)] * 2 * n,
            list(range(n)) * 3 + list(range(n, 2 * n)) * 3,
            [0] * n + list(range(n)) + list(range(0, -n, -1)) + [0] * n + list(range(n)) + list(range(0, -n, -1)),
        )
        self.n = n
        self.add_propagators([(list(range(i * n, i * n + n)), ALG_ALLDIFFERENT, []) for i in range(6)])
        self.add_propagator((list(range(n)) + list(range(3 * n, 4 * n)), ALG_PERMUTATION_AUX, []))

    def solution_as_printable(self, solution: NDArray) -> Any:
        return [([" "] * i + ["Q"] + [" "] * (self.n - i - 1)) for i in (solution[: self.n])]



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import sys

from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import MAX, MIN


@njit(cache=True)
def max_regret_var_heuristic(
    decision_domains: NDArray, domains_stk: NDArray, stks_top: NDArray, params: NDArray
) -> int:
    """
    Chooses the variable with the maximal regret (difference between the best and second-best value).
    :param decision_domains: the indices of a subset of the shared domains
    :param domains_stk: the stack of shared domains
    :param stks_top: the index of the top of the stacks as a Numpy array
    :param params: a two-dimensional (first dimension corresponds to variables, second to values) costs array
    :return: the index of the shared domain
    """
    best_score = 0
    best_idx = -1
    top = stks_top[0]
    for dom_idx in decision_domains:
        shr_domain = domains_stk[top, dom_idx]
        if 0 < shr_domain[MAX] - shr_domain[MIN]:
            score = regret(shr_domain, params[dom_idx])
            if best_score < score:
                best_idx = dom_idx
                best_score = score
    return best_idx


@njit(cache=True)
def regret(domain: NDArray, costs: NDArray) -> int:
    best_cost = sys.maxsize
    second_cost = sys.maxsize
    for value in range(domain[MIN], domain[MAX] + 1):
        cost = costs[value]
        if 0 < cost < best_cost:
            second_cost = best_cost
            best_cost = cost
        elif 0 < cost < second_cost:
            second_cost = cost
    return second_cost - best_cost



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import sys

from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import EVENT_MASK_MIN_MAX, MAX, MIN, PROP_CONSISTENCY, PROP_ENTAILMENT, PROP_INCONSISTENCY


def get_complexity_element_l_eq(n: int, parameters: NDArray) -> float:
    """
    Returns the time complexity of the propagator as a float.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: a float
    """
    return n


@njit(cache=True)
def get_triggers_element_l_eq(n: int, dom_idx: int, parameters: NDArray) -> int:
    """
    This propagator is triggered whenever there is a change in the domain of a variable.
    :param parameters: the parameters, unused here
    :return: an array of triggers
    """
    return EVENT_MASK_MIN_MAX


@njit(cache=True)
def compute_domains_element_l_eq(domains: NDArray, parameters: NDArray) -> int:
    """
    Enforces l_i = v.
    :param domains: the domains of the variables,
           l is the list of the first n-2 domains,
           i is the (n-1)th domain,
           v is the last domain
    :param parameters: the parameters of the propagator, it is unused
    :return: the status of the propagation (consistency, inconsistency or entailment) as an int
    """
    l = domains[:-2]
    i = domains[-2]
    v = domains[-1]
    # i could be updated only once
    i[MIN] = max(i[MIN], 0)
    i[MAX] = min(i[MAX], len(l) - 1)
    v_min = sys.maxsize
    v_max = -sys.maxsize
    start = -1
    for idx in range(i[MIN], i[MAX] + 1):
        if v[MAX] < l[idx, MIN] or v[MIN] > l[idx, MAX]:  # no intersection
            if start == -1:
                start = idx
            if idx == i[MIN]:
                i[MIN] += 1
        else:  # intersection
            start = -1
            if l[idx, MIN] < v_min:
                v_min = l[idx, MIN]
            if l[idx, MAX] > v_max:
                v_max = l[idx, MAX]
    if start >= 0:
        i[MAX] = start - 1
        if i[MAX] < i[MIN]:
            return PROP_INCONSISTENCY
    v[MIN] = max(v[MIN], v_min)
    v[MAX] = min(v[MAX], v_max)
    if i[MIN] == i[MAX]:
        l[i[MIN], :] = v[:]
        if v[MIN] == v[MAX]:
            return PROP_ENTAILMENT
    return PROP_CONSISTENCY



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import EVENT_MASK_MIN_MAX, MAX, MIN, PROP_CONSISTENCY, PROP_ENTAILMENT, PROP_INCONSISTENCY


def get_complexity_lexicographic_leq(n: int, parameters: NDArray) -> float:
    """
    Returns the time complexity of the propagator as a float.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: a float
    """
    return n


@njit(cache=True)
def get_triggers_lexicographic_leq(n: int, dom_idx: int, parameters: NDArray) -> int:
    """
    This propagator is triggered whenever there is a change in the domain of a variable.
    :param parameters: the parameters, unused here
    :return: an array of triggers
    """
    return EVENT_MASK_MIN_MAX


@njit(cache=True)
def compute_domains_4(x: NDArray, y: NDArray, n: int, i: int, q: int, r: int, s: int) -> int:
    while i < n and x[i, MIN] == y[i, MAX]:
        i += 1
        # s = i
    if i < n and x[i, MIN] > y[i, MAX]:
        # xq < yq
        x[q, MAX] = min(x[q, MAX], y[q, MAX] - 1)
        if x[q, MAX] < x[q, MIN]:
            return PROP_INCONSISTENCY
        y[q, MIN] = max(y[q, MIN], x[q, MIN] + 1)
        if y[q, MAX] < y[q, MIN]:
            return PROP_INCONSISTENCY
        return PROP_ENTAILMENT if x[q, MAX] < y[q, MIN] else PROP_CONSISTENCY
    # u = 4
    return PROP_CONSISTENCY


@njit(cache=True)
def compute_domains_3(x: NDArray, y: NDArray, n: int, i: int, q: int, r: int, s: int) -> int:
    while i < n and x[i, MAX] == y[i, MIN]:
        i += 1
        # s = i
    if i == n or x[i, MAX] < y[i, MIN]:
        # xq <= yq
        x[q, MAX] = min(x[q, MAX], y[q, MAX])
        if x[q, MAX] < x[q, MIN]:
            return PROP_INCONSISTENCY
        y[q, MIN] = max(y[q, MIN], x[q, MIN])
        if y[q, MAX] < y[q, MIN]:
            return PROP_INCONSISTENCY
        return PROP_ENTAILMENT if x[q, MAX] <= y[q, MIN] else PROP_CONSISTENCY
    # u = 3
    return PROP_CONSISTENCY


@njit(cache=True)
def compute_domains_2(x: NDArray, y: NDArray, n: int, i: int, q: int, r: int, s: int) -> int:
    while i < n and x[i, MIN] == x[i, MAX] == y[i, MIN] == y[i, MAX]:
        i += 1
        r = i
    if i == n or x[i, MAX] < y[i, MIN]:
        # xq <= yq
        x[q, MAX] = min(x[q, MAX], y[q, MAX])
        if x[q, MAX] < x[q, MIN]:
            return PROP_INCONSISTENCY
        y[q, MIN] = max(y[q, MIN], x[q, MIN])
        if y[q, MAX] < y[q, MIN]:
            return PROP_INCONSISTENCY
        return PROP_ENTAILMENT if x[q, MAX] <= y[q, MIN] else PROP_CONSISTENCY
    if x[i, MIN] > y[i, MAX]:
        # xq < yq
        x[q, MAX] = min(x[q, MAX], y[q, MAX] - 1)
        if x[q, MAX] < x[q, MIN]:
            return PROP_INCONSISTENCY
        y[q, MIN] = max(y[q, MIN], x[q, MIN] + 1)
        if y[q, MAX] < y[q, MIN]:
            return PROP_INCONSISTENCY
        return PROP_ENTAILMENT if x[q, MAX] < y[q, MIN] else PROP_CONSISTENCY
    if x[i, MAX] == y[i, MIN] and x[i, MIN] < y[i, MAX]:
        if s > i + 1:
            i = s
        else:
            i += 1
            s = i
        return compute_domains_3(x, y, n, i, q, r, s)
    if x[i, MIN] == y[i, MAX] and x[i, MAX] > y[i, MIN]:
        if s > i + 1:
            i = s
        else:
            i += 1
            s = i
        return compute_domains_4(x, y, n, i, q, r, s)
    # u = 2
    return PROP_CONSISTENCY


@njit(cache=True)
def compute_domains_1(x: NDArray, y: NDArray, n: int, i: int, q: int, r: int, s: int) -> int:
    while i < n and x[i, MIN] == y[i, MAX]:
        # enforce xi = yi
        x[i, MAX] = min(x[i, MAX], y[i, MAX])
        if x[i, MAX] < x[i, MIN]:
            return PROP_INCONSISTENCY
        y[i, MIN] = max(y[i, MIN], x[i, MIN])
        if y[i, MAX] < y[i, MIN]:
            return PROP_INCONSISTENCY
        i += 1
        q = i
    if i == n or x[i, MAX] < y[i, MIN]:
        return PROP_ENTAILMENT
    # enforce xq <= yq
    x[i, MAX] = min(x[i, MAX], y[i, MAX])
    if x[i, MAX] < x[i, MIN]:
        return PROP_INCONSISTENCY
    y[i, MIN] = max(y[i, MIN], x[i, MIN])
    if y[i, MAX] < y[i, MIN]:
        return PROP_INCONSISTENCY
    if r > i + 1:
        i = r
    else:
        i += 1
        r = i
    return compute_domains_2(x, y, n, i, q, r, s)


@njit(cache=True)
def compute_domains_lexicographic_leq(domains: NDArray, parameters: NDArray) -> int:
    """
    Implements lexicographic leq: x <_leq y.
    See https://www.diva-portal.org/smash/record.jsf?pid=diva2:1041533.
    :param domains: the domains of the variables,
           x is the list of the first n domains,
           y is the list of the last n domains
    :param parameters: unused here
    :return: the status of the propagation (consistency, inconsistency or entailment) as an int
    """
    # TODO: make incremental, use a var?
    n = len(domains) // 2
    return compute_domains_1(domains[:n], domains[n:], n, 0, 0, 0, 0)



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import os

from numba import bool, int32, int64, types, uint8, uint16, uint32, uint64  # type: ignore

# Optimizer modes
OPTIM_RESET = "RESET"
OPTIM_PRUNE = "PRUNE"
OPTIM_MODES = [OPTIM_RESET, OPTIM_PRUNE]


# Bounds
VARIABLE = 0  # index for a variable
PARAM = 1  # index for a parameter
RANGE_START = 0  # index corresponding to the start of a value range
RANGE_END = 1  # index corresponding to the end of a value range


# Domain bounds
MIN = 0  # min value of a domain
MAX = 1  # max value of a domain


# Domain update stack indices
DOM_UPDATE_IDX = 0  # index for the domain idx
DOM_UPDATE_EVENTS = 1  # index for the event


# Events
EVENT_NB = 3
EVENT_MASK_MIN = 1 << 0
EVENT_MASK_MAX = 1 << 1
EVENT_MASK_GROUND = 1 << 2
EVENT_MASK_MIN_MAX = EVENT_MASK_MIN | EVENT_MASK_MAX
EVENT_MASK_MIN_GROUND = EVENT_MASK_MIN | EVENT_MASK_GROUND
EVENT_MASK_MAX_GROUND = EVENT_MASK_MAX | EVENT_MASK_GROUND
EVENT_MASK_MIN_MAX_GROUND = EVENT_MASK_MIN | EVENT_MASK_MAX | EVENT_MASK_GROUND


PROP_INCONSISTENCY = 0  # returned by a propagator when inconsistent
PROP_CONSISTENCY = 1  # returned by a propagator when consistent
PROP_ENTAILMENT = 2  # returned by a propagator when entailed

PROBLEM_INCONSISTENT = 0  # returned when the filtering of a problem detects an inconsistency
PROBLEM_UNBOUND = 1  # returned when the filtering of a problem has been completed, but the problem is not solved
PROBLEM_BOUND = 2  # returned when a problem is solved

SIGNATURE_COMPUTE_DOMAINS = int64(int32[:, :], int32[:])  # domains, parameters
SIGNATURE_GET_TRIGGERS = int64(uint64, uint64, int32[:])
SIGNATURE_CONSISTENCY_ALG = int64(
    int64[:],  # statistics
    uint8[:],  # algorithms
    uint32[:, :, :],  # bounds
    uint32[:],  # variables_arr
    int32[:],  # offsets_arr
    uint32[:],  # props_variables
    int32[:],  # props_offsets
    int32[:],  # props_parameters
    int32[:, :, :],  # triggers
    int32[:, :, :],  # domains_stk
    bool[:, :],  # not_entailed_propagators_stk
    uint32[:, :],  # dom_update_stk
    uint16[:],  # stks_top
    uint32[:],  # triggered_propagators
    int64[:],  # compute_domains_addrs
    uint32[:],  # decision_domains
)
SIGNATURE_DOM_HEURISTIC = int64(
    int32[:, :, :],  # domains_stk
    bool[:, :],  # not_entailed_propagators_stk
    uint32[:, :],  # dom_update_stk
    uint16[:],  # stks_top
    int64,  # dom_idx
    int64[:, :],  # dom_heuristic_params
)
SIGNATURE_VAR_HEURISTIC = int64(
    uint32[:],  # decision_variables
    int32[:, :, :],  # domains_stk
    uint16[:],  # stks_top
    int64[:, :],  # var_heuristic_params
)


TYPE_COMPUTE_DOMAINS = types.FunctionType(SIGNATURE_COMPUTE_DOMAINS)
TYPE_GET_TRIGGERS = types.FunctionType(SIGNATURE_GET_TRIGGERS)
TYPE_DOM_HEURISTIC = types.FunctionType(SIGNATURE_DOM_HEURISTIC)
TYPE_VAR_HEURISTIC = types.FunctionType(SIGNATURE_VAR_HEURISTIC)
TYPE_CONSISTENCY_ALG = types.FunctionType(SIGNATURE_CONSISTENCY_ALG)

NUMBA_DISABLE_JIT = os.getenv("NUMBA_DISABLE_JIT")

LOG_FORMAT = "[ %(asctime)s | %(processName)s | %(levelname)s ] %(name)s.%(funcName)s - %(message)s"
LOG_LEVEL_DEBUG = "DEBUG"
LOG_LEVEL_INFO = "INFO"
LOG_LEVEL_WARNING = "WARNING"
LOG_LEVEL_ERROR = "ERROR"
LOG_LEVEL_CRITICAL = "CRITICAL"
LOG_LEVELS = [LOG_LEVEL_DEBUG, LOG_LEVEL_INFO, LOG_LEVEL_WARNING, LOG_LEVEL_ERROR, LOG_LEVEL_CRITICAL]

STATS_MAX = 16
(
    STATS_IDX_ALG_BC_NB,
    STATS_IDX_ALG_BC_WITH_SHAVING_NB,
    STATS_IDX_ALG_SHAVING_CHANGE_NB,
    STATS_IDX_ALG_SHAVING_NB,
    STATS_IDX_ALG_SHAVING_NO_CHANGE_NB,
    STATS_IDX_PROPAGATOR_ENTAILMENT_NB,
    STATS_IDX_PROPAGATOR_FILTER_NB,
    STATS_IDX_PROPAGATOR_FILTER_NO_CHANGE_NB,
    STATS_IDX_PROPAGATOR_INCONSISTENCY_NB,
    STATS_IDX_SEARCH_SPACE_INITIAL_SZ,
    STATS_IDX_SEARCH_SPACE_REMAINING_SZ,
    STATS_IDX_SEARCH_SPACE_LOG2_SCALE,
    STATS_IDX_SOLUTION_NB,
    STATS_IDX_SOLVER_BACKTRACK_NB,
    STATS_IDX_SOLVER_CHOICE_DEPTH,
    STATS_IDX_SOLVER_CHOICE_NB,
) = tuple(range(STATS_MAX))

STATS_LBL_ALG_BC_NB = "ALG_BC_NB"
STATS_LBL_ALG_BC_WITH_SHAVING_NB = "ALG_BC_WITH_SHAVING_NB"
STATS_LBL_ALG_SHAVING_CHANGE_NB = "ALG_SHAVING_CHANGE_NB"
STATS_LBL_ALG_SHAVING_NB = "ALG_SHAVING_NB"
STATS_LBL_ALG_SHAVING_NO_CHANGE_NB = "ALG_SHAVING_NO_CHANGE_NB"
STATS_LBL_PROPAGATOR_ENTAILMENT_NB = "PROPAGATOR_ENTAILMENT_NB"
STATS_LBL_PROPAGATOR_FILTER_NB = "PROPAGATOR_FILTER_NB"
STATS_LBL_PROPAGATOR_FILTER_NO_CHANGE_NB = "PROPAGATOR_FILTER_NO_CHANGE_NB"
STATS_LBL_PROPAGATOR_INCONSISTENCY_NB = "PROPAGATOR_INCONSISTENCY_NB"
STATS_LBL_SEARCH_SPACE_INITIAL_SZ = "SEARCH_SPACE_INITIAL_SZ"
STATS_LBL_SEARCH_SPACE_REMAINING_SZ = "SEARCH_SPACE_REMAINING_SZ"
STATS_LBL_SEARCH_SPACE_LOG2_SCALE = "SEARCH_SPACE_LOG2_SCALE"
STATS_LBL_SOLUTION_NB = "SOLUTION_NB"
STATS_LBL_SOLVER_BACKTRACK_NB = "SOLVER_BACKTRACK_NB"
STATS_LBL_SOLVER_CHOICE_DEPTH = "SOLVER_CHOICE_DEPTH"
STATS_LBL_SOLVER_CHOICE_NB = "SOLVER_CHOICE_NB"



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from typing import Any, List

from numpy.typing import NDArray

from nucs.problems.problem import Problem
from nucs.propagators.propagators import ALG_COUNT_EQ, ALG_COUNT_EQ_C, ALG_COUNT_LEQ_C


# TODO: GCC instead of booleans
# TODO: get rid of count variables, use count_leq and count_geq
# TODO: use custom heuristic (cf requests)
class EmployeeSchedulingProblem(Problem):
    """
    See https://developers.google.com/optimization/scheduling/employee_scheduling.
    """

    def shift_index(self, day: int, shift: int, nurse: int) -> int:
        return day * self.shift_nb * self.nurse_nb + shift * self.nurse_nb + nurse

    def nurses(self, day: int, shift: int) -> List[int]:
        start_shift = self.shift_index(day, shift, 0)
        return list(range(start_shift, start_shift + self.nurse_nb))

    def shifts(self, day: int, nurse: int) -> List[int]:
        start_shift = self.shift_index(day, 0, nurse)
        return list(
            range(
                start_shift,
                start_shift + self.shift_nb * self.nurse_nb,
                self.nurse_nb,
            )
        )

    def __init__(self) -> None:
        """
        Initializes the problem.
        """
        self.day_nb = 7
        self.shift_nb = 3
        self.nurse_nb = 5
        self.shift_total_nb = self.day_nb * self.shift_nb * self.nurse_nb
        self.shift_requests_nds = [
            [[0, 0, 1], [0, 0, 0], [0, 0, 0], [0, 0, 0], [0, 0, 1], [0, 1, 0], [0, 0, 1]],
            [[0, 0, 0], [0, 0, 0], [0, 1, 0], [0, 1, 0], [1, 0, 0], [0, 0, 0], [0, 0, 1]],
            [[0, 1, 0], [0, 1, 0], [0, 0, 0], [1, 0, 0], [0, 0, 0], [0, 1, 0], [0, 0, 0]],
            [[0, 0, 1], [0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 0], [1, 0, 0], [0, 0, 0]],
            [[0, 0, 0], [0, 0, 1], [0, 1, 0], [0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 0]],
        ]
        shift_requests_dsn = [
            self.shift_requests_nds[n][d][s]
            for d in range(self.day_nb)
            for s in range(self.shift_nb)
            for n in range(self.nurse_nb)
        ]
        min_shift_count_per_nurse = (self.shift_nb * self.day_nb) // self.nurse_nb
        max_shift_count_per_nurse = min_shift_count_per_nurse + (
            0 if self.shift_nb * self.day_nb % self.nurse_nb == 0 else 1
        )
        super().__init__([(0, 1)] * self.shift_total_nb)  # the boolean variables for shifts
        self.add_variables(
            [(min_shift_count_per_nurse, max_shift_count_per_nurse)] * self.nurse_nb
        )  # the number of shifts per nurse
        self.satisfied_request_nb = self.add_variable((0, self.shift_total_nb))  # the number of satisfied requests
        self.add_propagators(
            [(self.nurses(d, s), ALG_COUNT_EQ_C, [1, 1]) for d in range(self.day_nb) for s in range(self.shift_nb)]
        )
        self.add_propagators(
            [(self.shifts(d, n), ALG_COUNT_LEQ_C, [1, 1]) for d in range(self.day_nb) for n in range(self.nurse_nb)]
        )
        self.add_propagators(
            [
                (list(range(n, self.shift_total_nb + self.nurse_nb, self.nurse_nb)), ALG_COUNT_EQ, [1])
                for n in range(self.nurse_nb)
            ]
        )
        self.add_propagator(
            (
                [index for index in range(0, self.shift_total_nb) if shift_requests_dsn[index]]
                + [self.satisfied_request_nb],
                ALG_COUNT_EQ,
                [1],
            )
        )

    def solution_as_printable(self, solution: NDArray) -> Any:
        return {
            d: [
                [
                    (("W" if self.shift_requests_nds[n][d][s] else "w") if solution[self.shift_index(d, s, n)] else " ")
                    for n in range(self.nurse_nb)
                ]
                for s in range(self.shift_nb)
            ]
            for d in range(self.day_nb)
        }



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from typing import List, Optional, Tuple, Union

import pytest

from nucs.constants import PROP_CONSISTENCY, PROP_ENTAILMENT, PROP_INCONSISTENCY, STATS_IDX_SOLUTION_NB
from nucs.problems.problem import Problem
from nucs.propagators.lexicographic_leq_propagator import compute_domains_lexicographic_leq
from nucs.propagators.propagators import ALG_LEXICOGRAPHIC_LEQ
from nucs.solvers.backtrack_solver import BacktrackSolver
from tests.propagators.propagator_test import PropagatorTest


class TestLexicographicLeq(PropagatorTest):
    @pytest.mark.parametrize(
        "domains,parameters,consistency_result,expected_domains",
        [
            ([(0, 1), 0, 1, 1], [], PROP_ENTAILMENT, [[0, 1], [0, 0], [1, 1], [1, 1]]),
            ([(0, 1), (0, 1), (0, 1), (0, 1)], [], PROP_CONSISTENCY, [[0, 1], [0, 1], [0, 1], [0, 1]]),
            ([0, 0, 0, 0], [], PROP_ENTAILMENT, None),
            ([0, 0, 0, 1], [], PROP_ENTAILMENT, None),
            ([0, 0, 1, 0], [], PROP_ENTAILMENT, None),
            ([0, 0, 1, 1], [], PROP_ENTAILMENT, None),
            ([0, 1, 0, 0], [], PROP_INCONSISTENCY, None),
            ([0, 1, 0, 1], [], PROP_ENTAILMENT, None),
            ([0, 1, 1, 0], [], PROP_ENTAILMENT, None),
            ([0, 1, 1, 1], [], PROP_ENTAILMENT, None),
            ([1, 0, 0, 0], [], PROP_INCONSISTENCY, None),
            ([1, 0, 0, 1], [], PROP_INCONSISTENCY, None),
            ([1, 0, 1, 0], [], PROP_ENTAILMENT, None),
            ([1, 0, 1, 1], [], PROP_ENTAILMENT, None),
            ([1, 1, 0, 0], [], PROP_INCONSISTENCY, None),
            ([1, 1, 0, 1], [], PROP_INCONSISTENCY, None),
            ([1, 1, 1, 0], [], PROP_INCONSISTENCY, None),
            ([1, 1, 1, 1], [], PROP_ENTAILMENT, None),
        ],
    )
    def test_compute_domains(
        self,
        domains: List[Union[int, Tuple[int, int]]],
        parameters: List[int],
        consistency_result: int,
        expected_domains: Optional[List[List[int]]],
    ) -> None:
        self.assert_compute_domains(
            compute_domains_lexicographic_leq, domains, parameters, consistency_result, expected_domains
        )

    def test_solve_1(self) -> None:
        problem = Problem(
            domains=[(0, 1), (0, 1), (0, 1), (0, 1)],
            variables=[0, 1, 2, 3],
            offsets=[0, 0, 0, 0],
        )
        problem.add_propagator(([0, 1, 2, 3], ALG_LEXICOGRAPHIC_LEQ, []))
        solver = BacktrackSolver(problem)
        solver.solve_all()
        assert solver.statistics[STATS_IDX_SOLUTION_NB] == 10

    def test_solve_2(self) -> None:
        problem = Problem(
            domains=[(1, 1), (0, 1), (0, 1), (0, 1)],
            variables=[0, 1, 2, 3],
            offsets=[0, 0, 0, 0],
        )
        problem.add_propagator(([0, 1, 2, 3], ALG_LEXICOGRAPHIC_LEQ, []))
        solver = BacktrackSolver(problem)
        solver.solve_all()
        assert solver.statistics[STATS_IDX_SOLUTION_NB] == 3



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from typing import List, Optional, Tuple, Union

import pytest

from nucs.constants import PROP_CONSISTENCY, PROP_ENTAILMENT, PROP_INCONSISTENCY
from nucs.propagators.count_eq_c_propagator import compute_domains_count_eq_c
from tests.propagators.propagator_test import PropagatorTest


class TestCountEqC(PropagatorTest):
    @pytest.mark.parametrize(
        "domains,parameters,consistency_result,expected_domains",
        [
            (
                [(1, 4), (3, 5), (3, 6), (6, 8), 3, 5],
                [5, 1],
                PROP_CONSISTENCY,
                [[1, 4], [3, 4], [3, 6], [6, 8], [3, 3], [5, 5]],
            ),
            (
                [(1, 4), (3, 5), (3, 6), (6, 8), 3, 5],
                [5, 2],
                PROP_CONSISTENCY,
                [[1, 4], [3, 5], [3, 6], [6, 8], [3, 3], [5, 5]],
            ),
            ([(1, 4), (3, 5), (3, 6), (6, 8), 3, 5], [5, 0], PROP_INCONSISTENCY, None),
            ([2, (0, 1), (3, 4), 2, 2], [2, 3], PROP_ENTAILMENT, [[2, 2], [0, 1], [3, 4], [2, 2], [2, 2]]),
        ],
    )
    def test_compute_domains(
        self,
        domains: List[Union[int, Tuple[int, int]]],
        parameters: List[int],
        consistency_result: int,
        expected_domains: Optional[List[List[int]]],
    ) -> None:
        self.assert_compute_domains(
            compute_domains_count_eq_c, domains, parameters, consistency_result, expected_domains
        )



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from typing import List, Optional, Tuple, Union

import pytest

from nucs.constants import PROP_CONSISTENCY
from nucs.propagators.abs_eq_propagator import compute_domains_abs_eq
from tests.propagators.propagator_test import PropagatorTest


class TestAbsEq(PropagatorTest):
    @pytest.mark.parametrize(
        "domains,parameters,consistency_result,expected_domains",
        [
            ([(-4, 4), (-5, 5)], [], PROP_CONSISTENCY, [[-4, 4], [0, 4]]),
            ([(-4, 4), (-5, 0)], [], PROP_CONSISTENCY, [[0, 0], [0, 0]]),
            ([(2, 4), (-5, 5)], [], PROP_CONSISTENCY, [[2, 4], [2, 4]]),
            ([(-4, -2), (-5, 5)], [], PROP_CONSISTENCY, [[-4, -2], [2, 4]]),
        ],
    )
    def test_compute_domains(
        self,
        domains: List[Union[int, Tuple[int, int]]],
        parameters: List[int],
        consistency_result: int,
        expected_domains: Optional[List[List[int]]],
    ) -> None:
        self.assert_compute_domains(compute_domains_abs_eq, domains, parameters, consistency_result, expected_domains)



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import EVENT_MASK_MIN_MAX, MAX, MIN, PROP_CONSISTENCY, PROP_ENTAILMENT, PROP_INCONSISTENCY


def get_complexity_and_eq(n: int, parameters: NDArray) -> float:
    """
    Returns the time complexity of the propagator as a float.
    :param n: the number of variables
    :param parameters: the parameters, unused here
    :return: a float
    """
    return n


@njit(cache=True)
def get_triggers_and_eq(n: int, dom_idx: int, parameters: NDArray) -> int:
    """
    Returns the triggers for this propagator.
    :param parameters: the parameters, unused here
    :return: an array of triggers
    """
    return EVENT_MASK_MIN_MAX


@njit(cache=True)
def compute_domains_and_eq(domains: NDArray, parameters: NDArray) -> int:
    """
    Implements & b_i = b_{n-1} where for each i, b_i is a boolean variable.
    :param domains: the domains of the variables, b is an alias for domains
    :param parameters: unused here
    :return: the status of the propagation (consistency, inconsistency or entailment) as an int
    """
    x = domains[:-1]
    y = domains[-1]
    if np.any(x[:, MAX] == 0):
        y[MAX] = 0
        return PROP_INCONSISTENCY if y[MIN] > y[MAX] else PROP_ENTAILMENT
    if np.all(x[:, MIN] == 1):
        y[MIN] = 1
        return PROP_INCONSISTENCY if y[MIN] > y[MAX] else PROP_ENTAILMENT
    if y[MIN] == 1:
        x[:, MIN] = 1
        return PROP_ENTAILMENT  # no need to check for inconsistency
    elif y[MAX] == 0:
        candidates_nb = 0
        candidate_idx = -1
        for i in range(len(x)):
            if x[i, MIN] == 0:
                candidate_idx = i
                candidates_nb += 1
        if candidates_nb == 1:
            x[candidate_idx, MAX] = 0
            return PROP_ENTAILMENT  # no need to check for inconsistency
    return PROP_CONSISTENCY



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
import numpy as np
from numba import njit  # type: ignore
from numpy.typing import NDArray

from nucs.constants import (
    EVENT_MASK_GROUND,
    EVENT_MASK_MAX,
    EVENT_MASK_MIN,
    MAX,
    MIN,
    NUMBA_DISABLE_JIT,
    PARAM,
    PROBLEM_BOUND,
    PROBLEM_INCONSISTENT,
    PROBLEM_UNBOUND,
    PROP_ENTAILMENT,
    PROP_INCONSISTENCY,
    RANGE_END,
    RANGE_START,
    STATS_IDX_ALG_BC_NB,
    STATS_IDX_PROPAGATOR_ENTAILMENT_NB,
    STATS_IDX_PROPAGATOR_FILTER_NB,
    STATS_IDX_PROPAGATOR_FILTER_NO_CHANGE_NB,
    STATS_IDX_PROPAGATOR_INCONSISTENCY_NB,
    TYPE_COMPUTE_DOMAINS,
    VARIABLE,
)
from nucs.heaps import min_heap_pop
from nucs.numba_helper import function_from_address
from nucs.propagators.propagators import COMPUTE_DOMAINS_FCTS, update_propagators
from nucs.solvers.solver import is_solved


@njit(cache=True)
def bound_consistency_algorithm(
    statistics: NDArray,
    algorithms: NDArray,
    bounds: NDArray,
    variables_arr: NDArray,
    offsets_arr: NDArray,
    props_variables: NDArray,
    props_offsets: NDArray,
    props_parameters: NDArray,
    triggers: NDArray,
    domains_stk: NDArray,
    not_entailed_propagators_stk: NDArray,
    dom_update_stk: NDArray,
    stks_top: NDArray,
    triggered_propagators: NDArray,
    compute_domains_addrs: NDArray,
    decision_domains: NDArray,
) -> int:
    """
    Bound consistency algorithm.
    :param statistics: a Numpy array of statistics
    :param algorithms: the algorithms indexed by propagators
    :param bounds: the bounds indexed by propagators
    :param variables_arr: the domain indices indexed by variables, unused here
    :param offsets_arr: the domain offsets indexed by variables, unused here
    :param props_variables: the domain indices indexed by propagator variables
    :param props_offsets: the domain offsets indexed by propagator variables
    :param props_parameters: the parameters indexed by propagator variables
    :param triggers: a Numpy array of event masks indexed by shared domain indices and propagators
    :param domains_stk: a stack of shared domains;
    the first level correspond to the current shared domains, the rest correspond to the choice points
    :param not_entailed_propagators_stk: a stack not entailed propagators;
    the first level correspond to the propagators currently not entailed, the rest correspond to the choice points
    :param dom_update_stk: the stack of domain updates, unused here
    :param stks_top: the height of the stacks as a Numpy array
    :param triggered_propagators: the Numpy array of triggered propagators
    :param compute_domains_addrs: the addresses of the compute_domains functions
    :param decision_domains: the indices of the shared domains on which decisions will be made
    :return: a status (consistency, inconsistency or entailment) as an integer
    """
    top = stks_top[0]
    statistics[STATS_IDX_ALG_BC_NB] += 1
    propagator_nb = len(algorithms)
    while True:
        prop_idx = min_heap_pop(triggered_propagators, propagator_nb)
        if prop_idx == -1:
            return PROBLEM_BOUND if is_solved(domains_stk, top) else PROBLEM_UNBOUND
        statistics[STATS_IDX_PROPAGATOR_FILTER_NB] += 1
        prop_var_start = bounds[prop_idx, VARIABLE, RANGE_START]
        prop_var_end = bounds[prop_idx, VARIABLE, RANGE_END]
        prop_var_nb = prop_var_end - prop_var_start
        prop_domains = np.empty((prop_var_nb, 2), dtype=np.int32)
        for var_idx in range(prop_var_nb):
            prop_var_idx = prop_var_start + var_idx
            prop_dom_idx = props_variables[prop_var_idx]
            prop_offset = props_offsets[prop_var_idx]
            prop_domains[var_idx, MIN] = domains_stk[top, prop_dom_idx, MIN] + prop_offset
            prop_domains[var_idx, MAX] = domains_stk[top, prop_dom_idx, MAX] + prop_offset
        compute_domains_fct = (
            COMPUTE_DOMAINS_FCTS[algorithms[prop_idx]]
            if NUMBA_DISABLE_JIT
            else function_from_address(TYPE_COMPUTE_DOMAINS, compute_domains_addrs[algorithms[prop_idx]])
        )
        status = compute_domains_fct(
            prop_domains,
            props_parameters[bounds[prop_idx, PARAM, RANGE_START] : bounds[prop_idx, PARAM, RANGE_END]],
        )
        if status == PROP_INCONSISTENCY:
            statistics[STATS_IDX_PROPAGATOR_INCONSISTENCY_NB] += 1
            return PROBLEM_INCONSISTENT
        if status == PROP_ENTAILMENT:
            not_entailed_propagators_stk[top, prop_idx] = False
            statistics[STATS_IDX_PROPAGATOR_ENTAILMENT_NB] += 1
        no_changes = True
        for var_idx in range(prop_var_nb):
            prop_var_idx = prop_var_start + var_idx
            prop_dom_idx = props_variables[prop_var_idx]
            prop_offset = props_offsets[prop_var_idx]
            domain_min = prop_domains[var_idx, MIN] - prop_offset
            domain_max = prop_domains[var_idx, MAX] - prop_offset
            events = 0
            if domains_stk[top, prop_dom_idx, MIN] != domain_min:
                domains_stk[top, prop_dom_idx, MIN] = domain_min
                events |= EVENT_MASK_MIN
            if domains_stk[top, prop_dom_idx, MAX] != domain_max:
                domains_stk[top, prop_dom_idx, MAX] = domain_max
                events |= EVENT_MASK_MAX
            if events and domain_min == domain_max:
                events |= EVENT_MASK_GROUND
            if events:
                update_propagators(
                    triggered_propagators, not_entailed_propagators_stk[top], triggers, events, prop_dom_idx, prop_idx
                )
                no_changes = False
        if no_changes:
            statistics[STATS_IDX_PROPAGATOR_FILTER_NO_CHANGE_NB] += 1



###############################################################################
# __   _            _____    _____
# | \ | |          / ____|  / ____|
# |  \| |  _   _  | |      | (___
# | . ` | | | | | | |       \___ \
# | |\  | | |_| | | |____   ____) |
# |_| \_|  \__,_|  \_____| |_____/
#
# Fast constraint solving in Python  - https://github.com/yangeorget/nucs
#
# Copyright 2024-2025 - Yan Georget
###############################################################################
from nucs.constants import OPTIM_MODES, OPTIM_PRUNE
from nucs.examples.default_argument_parser import DefaultArgumentParser
from nucs.examples.employee_scheduling.employee_scheduling_problem import EmployeeSchedulingProblem
from nucs.heuristics.heuristics import DOM_HEURISTIC_MAX_VALUE
from nucs.solvers.backtrack_solver import BacktrackSolver

# Run with the following command (the second run is much faster because the code has been compiled):
# NUMBA_CACHE_DIR=.numba/cache python -m nucs.examples.employee_scheduling
if __name__ == "__main__":
    parser = DefaultArgumentParser()
    parser.add_argument("--opt_mode", choices=OPTIM_MODES, default=OPTIM_PRUNE)
    args = parser.parse_args()
    problem = EmployeeSchedulingProblem()
    solver = BacktrackSolver(problem, dom_heuristic_idx=DOM_HEURISTIC_MAX_VALUE, log_level=args.log_level)
    solution = solver.maximize(problem.satisfied_request_nb, mode=args.opt_mode)
    if args.stats:
        solver.print_statistics()
    if args.display:
        problem.print_solution(solution)
